"""Definitions of Netatmo devices types."""

from __future__ import annotations

from enum import StrEnum
import logging
from typing import Literal

LOG: logging.Logger = logging.getLogger(__name__)


class DeviceType(StrEnum):
    """Class to represent Netatmo device types."""

    # Climate/Energy
    NAPlug = "NAPlug"  # Smart thermostat gateway
    NATherm1 = "NATherm1"  # Smart thermostat
    NRV = "NRV"  # Smart valve
    NAC = "NAC"  # Smart AC control
    OTH = "OTH"  # OpenTherm gateway
    OTM = "OTM"  # OpenTherm modulating thermostat

    # Cameras/Security
    NACamDoorTag = "NACamDoorTag"  # Smart Door and Window Sensors
    NACamera = "NACamera"  # Smart Indoor Camera
    NCO = "NCO"  # Smart Carbon Monoxide Alarm
    NDB = "NDB"  # Smart Video Doorbell
    NIS = "NIS"  # Smart Indoor Siren
    NOC = "NOC"  # Smart Outdoor Camera (with Siren)
    NPC = "NPC"  # Indoor Camera Advance
    NSD = "NSD"  # Smart Smoke Detector

    # Weather
    NAMain = "NAMain"  # Smart Home Weather Station
    NAModule1 = "NAModule1"
    NAModule2 = "NAModule2"
    NAModule3 = "NAModule3"
    NAModule4 = "NAModule4"
    public = "public"

    # Home Coach
    NHC = "NHC"  # Smart Indoor Air Quality Monitor

    # Legrand Wiring devices and electrical panel products
    NLC = "NLC"  # Cable outlet
    NLD = "NLD"  # Remote control double on off dimmer
    NLDD = "NLDD"  # Dimmer
    NLE = "NLE"  # Connected Ecometer
    NLF = "NLF"  # Dimmer Light Switch
    NLFE = "NLFE"  # Dimmer Light Switch Evolution
    NLFN = "NLFN"  # light switch with neutral
    NLG = "NLG"  # Gateway
    NLGS = "NLGS"  # Gateway standalone
    NLIS = "NLIS"  # Double light switch
    NLL = "NLL"  # Italian light switch with neutral
    NLLM = "NLLM"  # Legrand / BTicino shutters
    NLLV = "NLLV"  # Legrand / BTicino shutters
    NLM = "NLM"  # light micro module
    NLP = "NLP"  # Plug
    NLPBS = "NLPBS"  # British standard plugs
    NLPC = "NLPC"  # Connected energy meter
    NLPD = "NLPD"  # Dry contact
    NLPM = "NLPM"  # mobile plug
    NLPO = "NLPO"  # Connected contactor
    NLPS = "NLPS"  # Smart Load Shedder
    NLPT = "NLPT"  # Connected latching relay / Telerupt
    NLT = "NLT"  # Remote control
    NLV = "NLV"  # Legrand / BTicino shutters
    NLAO = "NLAO"  # Legrand wireless batteryless light switch
    NLUO = "NLUO"  # Legrand Plug-In dimmer switch
    NLUI = "NLUI"  # Legrand In-Wall ON/OFF switch
    NLunknown = "NLunknown"  # Legrand device stub
    NLUF = "NLUF"  # Legrand device stub
    NLAS = "NLAS"  # Legrand wireless batteryless scene switch
    NLUP = "NLUP"  # Legrand device stub
    NLLF = "NLLF"  # Legrand Centralized Ventilation Control
    NLTS = "NLTS"  # Legrand motion sensor stub
    NLJ = "NLJ"  # Legrand garage door opener
    NLY = "NLY"  # Connected 3 phases energy meter

    # BTicino Classe 300 EOS
    BNCX = "BNCX"  # internal panel = gateway
    BNDL = "BNDL"  # door lock
    BNEU = "BNEU"  # external unit
    BNSL = "BNSL"  # staircase light
    BNCS = "BNCS"  # Controlled Socket
    BNXM = "BNXM"  # X meter
    BNMS = "BNMS"  # motorized shade
    BNAS = "BNAS"  # automatic shutter
    BNAB = "BNAB"  # automatic blind
    BNMH = "BNMH"  # MyHome server
    BNTH = "BNTH"  # thermostat
    BNFC = "BNFC"  # fan coil
    BNTR = "BNTR"  # radiator
    BNIL = "BNIL"  # intelligent light
    BNLD = "BNLD"  # dimmer light

    # Bubbendorf shutters
    NBG = "NBG"  # gateway
    NBO = "NBO"  # orientable shutter
    NBR = "NBR"  # roller shutter
    NBS = "NBS"  # swing shutter

    # VELUX ACTIVE
    NXG = "NXG"  # gateway
    NXO = "NXO"  # opener / cover

    # Somfy
    TPSRS = "TPSRS"  # Somfy io shutter

    # 3rd Party
    BNS = "BNS"  # Smarther with Netatmo
    EBU = "EBU"  # EBU gas meter
    Z3L = "Z3L"  # Zigbee 3 Light
    Z3V = "Z3V"  # Zigbee 3 roller shutter

    # Magellan
    NLDP = "NLDP"  # Pocket Remote

    @classmethod
    def _missing_(cls, key: object) -> Literal[DeviceType.NLunknown]:
        """Handle unknown device types."""

        msg = f"{key} device is unknown"
        LOG.warning(msg)
        return DeviceType.NLunknown


class DeviceCategory(StrEnum):
    """Class to represent Netatmo device types."""

    # temporarily disable locally-disabled and locally-enabled

    climate = "climate"
    camera = "camera"
    siren = "siren"
    shutter = "shutter"
    lock = "lock"
    switch = "switch"
    sensor = "sensor"
    weather = "weather"
    air_care = "air_care"
    meter = "meter"
    dimmer = "dimmer"
    opening = "opening"
    fan = "fan"


DEVICE_CATEGORY_MAP: dict[DeviceType, DeviceCategory] = {
    DeviceType.NRV: DeviceCategory.climate,
    DeviceType.NAC: DeviceCategory.climate,
    DeviceType.NATherm1: DeviceCategory.climate,
    DeviceType.OTM: DeviceCategory.climate,
    DeviceType.NOC: DeviceCategory.camera,
    DeviceType.NPC: DeviceCategory.camera,
    DeviceType.NACamDoorTag: DeviceCategory.opening,
    DeviceType.NACamera: DeviceCategory.camera,
    DeviceType.NDB: DeviceCategory.camera,
    DeviceType.NAMain: DeviceCategory.weather,
    DeviceType.NAModule1: DeviceCategory.weather,
    DeviceType.NAModule2: DeviceCategory.weather,
    DeviceType.NAModule3: DeviceCategory.weather,
    DeviceType.NAModule4: DeviceCategory.weather,
    DeviceType.NHC: DeviceCategory.air_care,
    DeviceType.NLV: DeviceCategory.shutter,
    DeviceType.NLLV: DeviceCategory.shutter,
    DeviceType.NLLM: DeviceCategory.shutter,
    DeviceType.NBR: DeviceCategory.shutter,
    DeviceType.NBO: DeviceCategory.shutter,
    DeviceType.NXO: DeviceCategory.shutter,
    DeviceType.NLP: DeviceCategory.switch,
    DeviceType.NLPM: DeviceCategory.switch,
    DeviceType.NLPBS: DeviceCategory.switch,
    DeviceType.NLIS: DeviceCategory.switch,
    DeviceType.NLL: DeviceCategory.switch,
    DeviceType.NLM: DeviceCategory.switch,
    DeviceType.NLC: DeviceCategory.switch,
    DeviceType.NLFN: DeviceCategory.dimmer,
    DeviceType.NLF: DeviceCategory.dimmer,
    DeviceType.NLFE: DeviceCategory.dimmer,
    DeviceType.BNS: DeviceCategory.climate,
    DeviceType.NLPC: DeviceCategory.meter,
    DeviceType.NLE: DeviceCategory.meter,
    DeviceType.Z3L: DeviceCategory.dimmer,
    DeviceType.Z3V: DeviceCategory.shutter,
    DeviceType.NLUP: DeviceCategory.switch,
    DeviceType.NLPO: DeviceCategory.switch,
    DeviceType.TPSRS: DeviceCategory.shutter,
    DeviceType.NLUO: DeviceCategory.dimmer,
    DeviceType.NLUI: DeviceCategory.switch,
    DeviceType.NLUF: DeviceCategory.dimmer,
    DeviceType.NLPS: DeviceCategory.meter,
    DeviceType.NLDD: DeviceCategory.switch,
    DeviceType.NLPT: DeviceCategory.switch,
    DeviceType.BNMS: DeviceCategory.shutter,
    DeviceType.BNAS: DeviceCategory.shutter,
    DeviceType.BNAB: DeviceCategory.shutter,
    DeviceType.BNTH: DeviceCategory.climate,
    DeviceType.BNFC: DeviceCategory.climate,
    DeviceType.BNTR: DeviceCategory.climate,
    DeviceType.NLPD: DeviceCategory.switch,
    DeviceType.NLJ: DeviceCategory.shutter,
    DeviceType.BNIL: DeviceCategory.switch,
    DeviceType.BNLD: DeviceCategory.dimmer,
    DeviceType.NIS: DeviceCategory.siren,
    DeviceType.BNCS: DeviceCategory.switch,
    DeviceType.NLLF: DeviceCategory.fan,
}


DEVICE_DESCRIPTION_MAP: dict[DeviceType, tuple[str, str]] = {
    # Netatmo Climate/Energy
    DeviceType.NAPlug: ("Netatmo", "Smart Thermostat Gateway"),
    DeviceType.NATherm1: ("Netatmo", "Smart Thermostat"),
    DeviceType.NRV: ("Netatmo", "Smart Valve"),
    DeviceType.NAC: ("Netatmo", "Smart AC Control"),
    DeviceType.OTH: ("Netatmo", "OpenTherm Gateway"),
    DeviceType.OTM: ("Netatmo", "OpenTherm Modulating Thermostat"),
    # Netatmo Cameras/Security
    DeviceType.NOC: ("Netatmo", "Smart Outdoor Camera"),
    DeviceType.NPC: ("Netatmo", "Indoor Camera Advance"),
    DeviceType.NACamera: ("Netatmo", "Smart Indoor Camera"),
    DeviceType.NSD: ("Netatmo", "Smart Smoke Detector"),
    DeviceType.NIS: ("Netatmo", "Smart Indoor Siren"),
    DeviceType.NACamDoorTag: ("Netatmo", "Smart Door/Window Sensors"),
    DeviceType.NDB: ("Netatmo", "Smart Video Doorbell"),
    DeviceType.NCO: ("Netatmo", "Smart Carbon Monoxide Alarm"),
    # Netatmo Weather
    DeviceType.NAMain: ("Netatmo", "Smart Home Weather station"),
    DeviceType.NAModule1: ("Netatmo", "Smart Outdoor Module"),
    DeviceType.NAModule2: ("Netatmo", "Smart Anemometer"),
    DeviceType.NAModule3: ("Netatmo", "Smart Rain Gauge"),
    DeviceType.NAModule4: ("Netatmo", "Smart Indoor Module"),
    DeviceType.public: ("Netatmo", "Public Weather station"),
    # Netatmo Home Coach
    DeviceType.NHC: ("Netatmo", "Smart Indoor Air Quality Monitor"),
    # Legrand Wiring devices and electrical panel products
    DeviceType.NLG: ("Legrand", "Gateway"),
    DeviceType.NLGS: ("Legrand", "Gateway standalone"),
    DeviceType.NLP: ("Legrand", "Plug"),
    DeviceType.NLPM: ("Legrand", "Mobile plug"),
    DeviceType.NLPBS: ("Legrand", "British standard plugs"),
    DeviceType.NLF: ("Legrand", "2 wire light switch/dimmer"),
    DeviceType.NLFE: ("Legrand", "2 wire light switch/dimmer evolution"),
    DeviceType.NLIS: ("Legrand", "Double switch"),
    DeviceType.NLFN: ("Legrand", "Light switch/dimmer with neutral"),
    DeviceType.NLM: ("Legrand", "Light micro module"),
    DeviceType.NLL: ("Legrand", "Italian light switch with neutral"),
    DeviceType.NLLF: ("Legrand", "Centralized ventilation device"),
    DeviceType.NLV: ("Legrand/BTicino", "Shutters"),
    DeviceType.NLLV: ("Legrand/BTicino", "Shutters"),
    DeviceType.NLLM: ("Legrand/BTicino", "Shutters"),
    DeviceType.NLPO: ("Legrand", "Connected Contactor"),
    DeviceType.NLPT: ("Legrand", "Connected Latching Relay"),
    DeviceType.NLPC: ("Legrand", "Connected Energy Meter"),
    DeviceType.NLE: ("Legrand", "Connected Ecometer"),
    DeviceType.NLPS: ("Legrand", "Smart Load Shedder"),
    DeviceType.NLC: ("Legrand", "Cable Outlet"),
    DeviceType.NLT: ("Legrand", "Global Remote Control"),
    DeviceType.NLAS: ("Legrand", "Wireless batteryless scene switch"),
    DeviceType.NLD: ("Legrand", "Remote Control"),
    DeviceType.NLDD: ("Legrand", "Dimmer"),
    DeviceType.NLUP: ("Legrand", "Power outlet"),
    DeviceType.NLUO: ("Legrand", "Plug-In dimmer switch"),
    DeviceType.NLUI: ("Legrand", "In-wall switch"),
    DeviceType.NLTS: ("Legrand", "Motion sensor"),
    DeviceType.NLUF: ("Legrand", "In-Wall dimmer"),
    DeviceType.NLJ: ("Legrand", "Garage door opener"),
    # BTicino Classe 300 EOS
    DeviceType.BNCX: ("BTicino", "Internal Panel"),
    DeviceType.BNEU: ("BTicino", "External Unit"),
    DeviceType.BNDL: ("BTicino", "Door Lock"),
    DeviceType.BNSL: ("BTicino", "Staircase Light"),
    DeviceType.BNMS: ("BTicino", "Motorized Shade"),
    DeviceType.BNAS: ("BTicino", "Automatic Shutter"),
    DeviceType.BNAB: ("BTicino", "Automatic Blind"),
    DeviceType.BNMH: ("BTicino", "MyHome server 1"),
    DeviceType.BNTH: ("BTicino", "Thermostat"),
    DeviceType.BNFC: ("BTicino", "Fan coil"),
    DeviceType.BNTR: ("BTicino", "Module towel rail"),
    DeviceType.BNIL: ("BTicino", "Intelligent light"),
    DeviceType.BNLD: ("BTicino", "Dimmer"),
    DeviceType.BNCS: ("BTicino", "Controlled socket"),
    # Bubbendorf shutters
    DeviceType.NBG: ("Bubbendorf", "Gateway"),
    DeviceType.NBR: ("Bubbendorf", "Roller Shutter"),
    DeviceType.NBO: ("Bubbendorf", "Orientable Shutter"),
    DeviceType.NBS: ("Bubbendorf", "Swing Shutter"),
    # VELUX ACTIVE
    DeviceType.NXG: ("VELUX ACTIVE", "Gateway"),
    DeviceType.NXO: ("VELUX ACTIVE", "Opener"),
    # Somfy
    DeviceType.TPSRS: ("Somfy", "io Shutter"),
    # 3rd Party
    DeviceType.BNS: ("Smarther", "Smarther with Netatmo"),
    DeviceType.Z3L: ("3rd Party", "Zigbee 3 Light"),
    DeviceType.Z3V: ("3rd Party", "Zigbee 3 roller shutter"),
    DeviceType.EBU: ("3rd Party", "EBU gas meter"),
    DeviceType.NLPD: ("Drivia", "Dry contact"),
}


class ApplianceType(StrEnum):
    """Class to represent appliance type of a module. This is only for Home + Control."""

    # temporarily disable locally-disabled and locally-enabled

    light = "light"
    fridge_freezer = "fridge_freezer"
    oven = "oven"
    washing_machine = "washing_machine"
    tumble_dryer = "tumble_dryer"
    dishwasher = "dishwasher"
    multimedia = "multimedia"
    router = "router"
    other = "other"
    cooking = "cooking"
    radiator = "radiator"
    radiator_without_pilot_wire = "radiator_without_pilot_wire"
    water_heater = "water_heater"
    extractor_hood = "extractor_hood"
    contactor = "contactor"
    dryer = "dryer"
    electric_charger = "electric_charger"
    unknown = "unknown"

    @classmethod
    def _missing_(cls, key: object) -> Literal[ApplianceType.unknown]:
        """Handle unknown device types."""

        msg: str = f"{key} appliance type is unknown"
        LOG.warning(msg)
        return ApplianceType.unknown


class DoorTagCategory(StrEnum):
    """Class to represent category of a module. This is only for Home + Security/NACamDoorTag."""

    door = "door"
    furniture = "furniture"
    garage = "garage"
    gate = "gate"
    other = "other"
    window = "window"
    unknown = "unknown"

    @classmethod
    def _missing_(cls, key: object) -> Literal[DoorTagCategory.unknown]:
        """Handle unknown device category."""

        msg: str = f"{key} category is unknown"
        LOG.warning(msg)
        return DoorTagCategory.unknown


class BoilerControl(StrEnum):
    """Boiler control mode reported by an OpenTherm gateway (OTH)."""

    onoff = "onoff"
    opentherm = "opentherm"
    detecting = "detecting"
    unknown = "unknown"

    @classmethod
    def _missing_(cls, key: object) -> Literal[BoilerControl.unknown]:
        """Handle unknown boiler control values."""

        msg: str = f"{key} boiler control is unknown"
        LOG.warning(msg)
        return BoilerControl.unknown


class BoilerError(StrEnum):
    """Boiler error reported by an OpenTherm gateway (OTH)."""

    boiler_not_responding = "boiler_not_responding"
    maintenance = "maintenance"
    water_pressure = "water_pressure"
    boiler_flame = "boiler_flame"
    air_pressure = "air_pressure"
    boiler_temperature = "boiler_temperature"
    unknown = "unknown"

    @classmethod
    def _missing_(cls, key: object) -> Literal[BoilerError.unknown]:
        """Handle unknown boiler error values."""

        msg: str = f"{key} boiler error is unknown"
        LOG.warning(msg)
        return BoilerError.unknown


class DhwControl(StrEnum):
    """Domestic-hot-water control reported by an OpenTherm gateway (OTH)."""

    none = "none"
    instantaneous = "instantaneous"
    water_tank = "water_tank"
    unknown = "unknown"

    @classmethod
    def _missing_(cls, key: object) -> Literal[DhwControl.unknown]:
        """Handle unknown domestic-hot-water control values."""

        msg: str = f"{key} dhw control is unknown"
        LOG.warning(msg)
        return DhwControl.unknown



"""Module to represent iDiamant modules."""

from __future__ import annotations

import logging

from pyatmo.modules.module import (
    FirmwareMixin,
    Module,
    RfMixin,
    ShutterMixin,
    WifiMixin,
)

LOG: logging.Logger = logging.getLogger(__name__)


class NBG(FirmwareMixin, WifiMixin, Module):
    """Class to represent a iDiamant NBG."""


class NBR(FirmwareMixin, RfMixin, ShutterMixin, Module):
    """Class to represent a iDiamant NBR."""


class NBO(FirmwareMixin, RfMixin, ShutterMixin, Module):
    """Class to represent a iDiamant NBO."""


class NBS(FirmwareMixin, RfMixin, ShutterMixin, Module):
    """Class to represent a iDiamant NBS."""



"""Enums for the pyatmo package."""

from __future__ import annotations

from enum import IntEnum, StrEnum
import logging
from typing import Literal

LOG: logging.Logger = logging.getLogger(__name__)


class UnitSystem(IntEnum):
    """Measurement system reported in the /homesdata user block."""

    METRIC = 0
    IMPERIAL = 1
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> Literal[UnitSystem.UNKNOWN]:
        """Handle unknown unit system values."""

        msg: str = f"{value} unit system is unknown"
        LOG.warning(msg)
        return UnitSystem.UNKNOWN


class WindUnit(IntEnum):
    """Wind-speed unit reported in the /homesdata user block."""

    KPH = 0
    MPH = 1
    MS = 2
    BEAUFORT = 3
    KNOT = 4
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> Literal[WindUnit.UNKNOWN]:
        """Handle unknown wind unit values."""

        msg: str = f"{value} wind unit is unknown"
        LOG.warning(msg)
        return WindUnit.UNKNOWN


class PressureUnit(IntEnum):
    """Pressure unit reported in the /homesdata user block."""

    MBAR = 0
    INHG = 1
    MMHG = 2
    UNKNOWN = -1

    @classmethod
    def _missing_(cls, value: object) -> Literal[PressureUnit.UNKNOWN]:
        """Handle unknown pressure unit values."""

        msg: str = f"{value} pressure unit is unknown"
        LOG.warning(msg)
        return PressureUnit.UNKNOWN


class ScheduleType(StrEnum):
    """Enum representing the type of a schedule."""

    THERM = "therm"
    COOLING = "cooling"
    ELECTRICITY = "electricity"
    ELECTRICITY_PRODUCTION = "electricity_production"
    EVENT = "event"
    AUTO = "auto"
    ALGO = "algo"


class TemperatureControlMode(StrEnum):
    """Temperature control mode."""

    HEATING = "heating"
    COOLING = "cooling"
    AUTO = "auto"


SCHEDULE_TYPE_MAPPING: dict[TemperatureControlMode, ScheduleType] = {
    TemperatureControlMode.HEATING: ScheduleType.THERM,
    TemperatureControlMode.COOLING: ScheduleType.COOLING,
    TemperatureControlMode.AUTO: ScheduleType.AUTO,
}



"""Module to represent BTicino modules."""

from __future__ import annotations

import logging

from pyatmo.modules.module import (
    DimmableMixin,
    Module,
    Shutter,
    ShutterMixin,
    Switch,
    SwitchMixin,
)

LOG: logging.Logger = logging.getLogger(__name__)


class BNDL(Module):
    """BTicino door lock."""


class BNSL(Switch):
    """BTicino staircase light."""


class BNCX(Module):
    """BTicino internal panel = gateway."""


class BNEU(Module):
    """BTicino external unit."""


class BNCS(Switch):
    """Bticino module Controlled Socket."""


class BNXM(Module):
    """BTicino X meter."""


class BNMS(Shutter):
    """BTicino motorized shade."""


class BNAS(ShutterMixin, Module):
    """BTicino automatic shutter."""


class BNAB(Shutter):
    """BTicino automatic blind."""


class BNMH(Module):
    """BTicino MyHome server."""


class BNTH(Module):
    """BTicino thermostat."""


class BNFC(Module):
    """BTicino fan coil."""


class BNTR(Module):
    """BTicino radiator thermostat."""


class BNIL(SwitchMixin, Module):
    """BTicino intelligent light."""


class BNLD(DimmableMixin, SwitchMixin, Module):
    """BTicino dimmer light."""



"""Expose submodules."""

from pyatmo import const, modules
from pyatmo.account import AsyncAccount
from pyatmo.auth import AbstractAsyncAuth
from pyatmo.const import SIREN_BASE_URL
from pyatmo.exceptions import (
    ApiError,
    ApiHomeReachabilityError,
    ApiThrottlingError,
    InvalidHomeError,
    InvalidRoomError,
    InvalidScheduleError,
    NoDeviceError,
    NoScheduleError,
)
from pyatmo.home import Home
from pyatmo.modules import Module
from pyatmo.modules.device_types import DeviceType
from pyatmo.room import Room
from pyatmo.webrtc import WebRTCStream

__all__: list[str] = [
    "SIREN_BASE_URL",
    "AbstractAsyncAuth",
    "ApiError",
    "ApiHomeReachabilityError",
    "ApiThrottlingError",
    "AsyncAccount",
    "DeviceType",
    "Home",
    "InvalidHomeError",
    "InvalidRoomError",
    "InvalidScheduleError",
    "Module",
    "NoDeviceError",
    "NoScheduleError",
    "Room",
    "WebRTCStream",
    "const",
    "modules",
]



"""Module to represent a Netatmo home."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

from pyatmo import modules
from pyatmo.const import (
    EVENTS,
    SCHEDULES,
    SETPERSONSAWAY_ENDPOINT,
    SETPERSONSHOME_ENDPOINT,
    SETSTATE_ENDPOINT,
    SETTHERMMODE_ENDPOINT,
    SWITCHHOMESCHEDULE_ENDPOINT,
    SYNCHOMESCHEDULE_ENDPOINT,
    RawData,
)
from pyatmo.enums import SCHEDULE_TYPE_MAPPING, TemperatureControlMode
from pyatmo.event import Event
from pyatmo.exceptions import (
    ApiHomeReachabilityError,
    InvalidScheduleError,
    InvalidStateError,
    NoScheduleError,
)
from pyatmo.person import Person
from pyatmo.room import Room
from pyatmo.schedule import Schedule

if TYPE_CHECKING:
    from aiohttp import ClientResponse

    from pyatmo.auth import AbstractAsyncAuth
    from pyatmo.modules import Module
    from pyatmo.modules.netatmo import NACamera

LOG: logging.Logger = logging.getLogger(__name__)

# Legacy/typo module type strings some /homesdata schema variants document,
# mapped to the canonical type the library implements. Defensive: the live API
# is expected to send the canonical spelling.
MODULE_TYPE_ALIASES: dict[str, str] = {
    "NBD": "NDB",  # transposition of Smart Video Doorbell
    "NADoorTag": "NACamDoorTag",  # legacy Smart Door/Window Sensor name
}


class Home:
    """Class to represent a Netatmo home."""

    auth: AbstractAsyncAuth
    entity_id: str
    name: str
    altitude: int | None = None
    coordinates: list[float] | None = None
    country: str | None = None
    timezone: str | None = None
    rooms: dict[str, Room]
    modules: dict[str, Module]
    schedules: dict[str, Schedule]
    persons: dict[str, Person]
    events: dict[str, Event]

    temperature_control_mode: TemperatureControlMode | None = None
    therm_mode: str | None = None
    therm_setpoint_default_duration: int | None = None
    cooling_mode: str | None = None

    def __init__(self, auth: AbstractAsyncAuth, raw_data: RawData) -> None:
        """Initialize a Netatmo home instance."""

        self.auth = auth
        self.entity_id = raw_data["id"]
        self.name = raw_data.get("name", "Unknown")
        self.altitude = raw_data.get("altitude")
        self.coordinates = raw_data.get("coordinates")
        self.country = raw_data.get("country")
        self.timezone = raw_data.get("timezone")
        self.modules = {
            module["id"]: self.get_module(module)
            for module in raw_data.get("modules", [])
        }
        self.rooms = {
            room["id"]: Room(
                home=self,
                room=room,
                all_modules=self.modules,
            )
            for room in raw_data.get("rooms", [])
        }
        self.schedules = {
            s["id"]: Schedule(home=self, raw_data=s)
            for s in raw_data.get(SCHEDULES, [])
        }
        self.persons = {
            s["id"]: Person(home=self, raw_data=s) for s in raw_data.get("persons", [])
        }
        self.events = {}

        self.temperature_control_mode = get_temperature_control_mode(
            raw_data.get("temperature_control_mode"),
        )
        self.therm_mode = raw_data.get("therm_mode")
        self.therm_setpoint_default_duration = raw_data.get(
            "therm_setpoint_default_duration",
        )
        self.cooling_mode = raw_data.get("cooling_mode")

    def get_module(self, module: dict) -> Module:
        """Return module."""

        module_type = MODULE_TYPE_ALIASES.get(module["type"], module["type"])
        if module_type != module["type"]:
            LOG.debug("Aliased device type %s -> %s", module["type"], module_type)
            # Normalize so both the class lookup and DeviceType(...) in
            # Module.__init__ see the canonical type. Copy (not mutate) to avoid
            # rewriting the shared raw_data dict, which consumers may serialize.
            module = {**module, "type": module_type}

        try:
            return getattr(modules, module_type)(
                home=self,
                module=module,
            )
        except AttributeError:
            LOG.info("Unknown device type %s", module_type)
            return modules.NLunknown(
                home=self,
                module=module,
            )

    def update_topology(self, raw_data: RawData) -> None:
        """Update topology."""

        self.name = raw_data.get("name", "Unknown")
        # Geolocation is treated as sticky, unlike the live state below (name,
        # therm mode, ...): it is effectively static per home and only carried
        # in a full /homesdata payload, so a topology update that omits these
        # keys keeps the previously populated values instead of wiping them.
        self.altitude = raw_data.get("altitude", self.altitude)
        self.coordinates = raw_data.get("coordinates", self.coordinates)
        self.country = raw_data.get("country", self.country)
        self.timezone = raw_data.get("timezone", self.timezone)

        raw_modules = raw_data.get("modules", [])

        self.temperature_control_mode = get_temperature_control_mode(
            raw_data.get("temperature_control_mode"),
        )
        self.therm_mode = raw_data.get("therm_mode")
        self.therm_setpoint_default_duration = raw_data.get(
            "therm_setpoint_default_duration",
        )
        self.cooling_mode = raw_data.get("cooling_mode")

        for module in raw_modules:
            if (module_id := module["id"]) not in self.modules:
                self.modules[module_id] = self.get_module(module)
            else:
                self.modules[module_id].update_topology(module)

        # Drop module if has been removed
        for module in self.modules.keys() - {m["id"] for m in raw_modules}:
            self.modules.pop(module)

        raw_rooms = raw_data.get("rooms", [])
        for room in raw_rooms:
            if (room_id := room["id"]) not in self.rooms:
                self.rooms[room_id] = Room(
                    home=self,
                    room=room,
                    all_modules=self.modules,
                )
            else:
                self.rooms[room_id].update_topology(room)

        # Drop room if has been removed
        for room in self.rooms.keys() - {m["id"] for m in raw_rooms}:
            self.rooms.pop(room)

        raw_schedules = raw_data.get("schedules", [])
        for schedule in raw_schedules:
            if (schedule_id := schedule["id"]) not in self.schedules:
                self.schedules[schedule_id] = Schedule(home=self, raw_data=schedule)
            else:
                self.schedules[schedule_id].update_topology(schedule)

        # Drop schedule if has been removed
        for schedule in self.schedules.keys() - {s["id"] for s in raw_schedules}:
            self.schedules.pop(schedule)

    async def update(
        self,
        raw_data: RawData,
        do_raise_for_reachability_error: bool = False,
    ) -> None:
        """Update home with the latest data."""
        has_error = False
        for module in raw_data.get("errors", []):
            has_error = True
            module_id = module["id"]
            if module_id in self.modules:
                await self.modules[module_id].update({})
                # Set error_code AFTER update({}): update() reruns reflection
                # (_update_attributes) which would otherwise reset it to None.
                self.modules[module_id].error_code = module.get("code")
            else:
                LOG.warning(
                    "Error reported for unknown module id (%s); skipping",
                    module_id,
                )

        data = raw_data["home"]

        has_an_update = False
        for module in data.get("modules", []):
            has_an_update = True
            if module["id"] not in self.modules:
                # Register the newly-seen module directly. Routing through
                # update_topology with a partial `{"modules": [...]}` payload
                # would wipe home-level fields (name, therm state, geolocation)
                # whose keys are absent from this /homestatus data.
                self.modules[module["id"]] = self.get_module(module)
            await self.modules[module["id"]].update(module)
            # Clear any error code from a previous /homestatus errors[] entry:
            # the module is reported healthy again. Reflection in update() would
            # otherwise carry the stale code forward (raw data has no error_code).
            self.modules[module["id"]].error_code = None

        for room in data.get("rooms", []):
            has_an_update = True
            if room["id"] in self.rooms:
                self.rooms[room["id"]].update(room)
            else:
                LOG.warning(
                    "Room id (%s) not found in known rooms. Known room ids: %s (count=%d)",
                    room["id"],
                    list(self.rooms.keys()),
                    len(self.rooms),
                )

        for person_status in data.get("persons", []):
            # if there is a person update, it means the house has been updated
            has_an_update = True
            if person := self.persons.get(person_status["id"]):
                person.update(person_status)

        self.events = {
            s["id"]: Event(home_id=self.entity_id, raw_data=s)
            for s in data.get(EVENTS, [])
        }
        if len(self.events) > 0:
            has_an_update = True

        has_one_module_reachable = False
        for module in self.modules.values():
            if module.reachable:
                has_one_module_reachable = True
            if hasattr(module, "events"):
                module = cast("NACamera", module)
                module.events = [
                    event
                    for event in self.events.values()
                    if event.module_id == module.entity_id
                ]

        if (
            do_raise_for_reachability_error
            and has_error
            and has_one_module_reachable is False
            and has_an_update is False
        ):
            msg = "No Home update could be performed, all modules unreachable and not updated"
            raise ApiHomeReachabilityError(
                msg,
            )

    def get_selected_schedule(self) -> Schedule | None:
        """Return selected schedule for given home."""

        return next(
            (
                schedule
                for schedule in self.schedules.values()
                if schedule.selected
                and self.temperature_control_mode
                and schedule.type
                == SCHEDULE_TYPE_MAPPING[self.temperature_control_mode]
            ),
            None,
        )

    def get_available_schedules(self) -> list[Schedule]:
        """Return available schedules for given home."""

        return [
            schedule
            for schedule in self.schedules.values()
            if self.temperature_control_mode
            and schedule.type == SCHEDULE_TYPE_MAPPING[self.temperature_control_mode]
        ]

    def is_valid_schedule(self, schedule_id: str) -> bool:
        """Check if valid schedule."""

        return schedule_id in self.schedules

    def has_otm(self) -> bool:
        """Check if any room has an OTM device."""

        return any("OTM" in room.device_types for room in self.rooms.values())

    def has_bns(self) -> bool:
        """Check if any room has a BNS device."""

        return any("BNS" in room.device_types for room in self.rooms.values())

    def get_hg_temp(self) -> float | None:
        """Return frost guard temperature value for given home."""

        if (schedule := self.get_selected_schedule()) is None:
            return None
        return schedule.hg_temp

    def get_away_temp(self) -> float | None:
        """Return configured away temperature value for given home."""

        if (schedule := self.get_selected_schedule()) is None:
            return None
        return schedule.away_temp

    async def async_set_thermmode(
        self,
        mode: str | None,
        end_time: int | None = None,
        schedule_id: str | None = None,
    ) -> bool:
        """Set thermotat mode."""
        if schedule_id is not None and not self.is_valid_schedule(schedule_id):
            msg = f"{schedule_id} is not a valid schedule id."
            raise NoScheduleError(msg)
        if mode is None:
            msg = f"{mode} is not a valid mode."
            raise NoScheduleError(msg)

        post_params: dict[str, str] = {"home_id": self.entity_id, "mode": mode}
        if end_time is not None and mode in {"hg", "away"}:
            post_params["endtime"] = str(end_time)
        if schedule_id is not None and mode == "schedule":
            post_params["schedule_id"] = schedule_id
        LOG.debug(
            "Setting home (%s) mode to %s (%s)",
            self.entity_id,
            mode,
            schedule_id,
        )

        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=SETTHERMMODE_ENDPOINT,
            params=post_params,
        )

        return (await resp.json()).get("status") == "ok"

    async def async_switch_schedule(self, schedule_id: str) -> bool:
        """Switch the schedule."""
        if not self.is_valid_schedule(schedule_id):
            msg: str = f"{schedule_id} is not a valid schedule id"
            raise NoScheduleError(msg)
        LOG.debug("Setting home (%s) schedule to %s", self.entity_id, schedule_id)
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=SWITCHHOMESCHEDULE_ENDPOINT,
            params={"home_id": self.entity_id, "schedule_id": schedule_id},
        )

        return (await resp.json()).get("status") == "ok"

    async def async_set_state(self, data: dict[str, Any]) -> bool:
        """Set state using given data."""
        if not is_valid_state(data):
            msg = "Data for '/set_state' contains errors."
            raise InvalidStateError(msg)
        LOG.debug("Setting state for home (%s) according to %s", self.entity_id, data)
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=SETSTATE_ENDPOINT,
            params={"json": {"home": {"id": self.entity_id, **data}}},
        )

        response = await resp.json()
        body = response.get("body")
        if isinstance(body, dict) and (errors := body.get("errors")):
            LOG.warning(
                "Set state response for home %s contains errors: status=%r errors=%r",
                self.entity_id,
                response.get("status"),
                errors,
            )
            return False
        return response.get("status") == "ok"

    async def async_set_persons_home(
        self,
        person_ids: list[str] | None = None,
    ) -> ClientResponse:
        """Mark persons as home."""
        post_params: dict[str, Any] = {"home_id": self.entity_id}
        if person_ids:
            post_params["person_ids[]"] = person_ids
        return await self.auth.async_post_api_request(
            endpoint=SETPERSONSHOME_ENDPOINT,
            params=post_params,
        )

    async def async_set_persons_away(
        self,
        person_id: str | None = None,
    ) -> ClientResponse:
        """Mark a person as away or set the whole home to being empty."""

        post_params: dict[str, str] = {"home_id": self.entity_id}
        if person_id:
            post_params["person_id"] = person_id
        return await self.auth.async_post_api_request(
            endpoint=SETPERSONSAWAY_ENDPOINT,
            params=post_params,
        )

    async def async_set_schedule_temperatures(
        self,
        zone_id: str,
        temps: dict[str, float],
    ) -> None:
        """Set the scheduled room temperature for the given schedule ID."""

        selected_schedule: Schedule | None = self.get_selected_schedule()

        if selected_schedule is None:
            msg = "Could not determine selected schedule."
            raise NoScheduleError(msg)

        for zone in (z for z in selected_schedule.zones if z.entity_id == zone_id):
            for room in (r for r in zone.rooms if r.entity_id in temps):
                room.therm_setpoint_temperature = temps[room.entity_id]

        await self.async_sync_schedule(selected_schedule)

    async def async_sync_schedule(
        self,
        schedule: Schedule,
    ) -> None:
        """Modify an existing schedule."""
        if not is_valid_schedule(schedule):
            msg = "Data for '/synchomeschedule' contains errors."
            raise InvalidScheduleError(msg)
        LOG.debug(
            "Setting schedule (%s) for home (%s) to %s",
            schedule.entity_id,
            self.entity_id,
            schedule,
        )

        timetable_entries: list[dict[str, int | None]] = [
            {
                "m_offset": timetable_entry.m_offset,
                "zone_id": timetable_entry.zone_id,
            }
            for timetable_entry in schedule.timetable
        ]

        zones: list[dict[str, Any]] = []
        for zone in schedule.zones:
            new_zone: dict[str, Any] = {
                "id": zone.entity_id,
                "name": zone.name,
                "type": zone.type,
                "rooms": [
                    {
                        "id": room.entity_id,
                        "therm_setpoint_temperature": room.therm_setpoint_temperature,
                    }
                    for room in zone.rooms
                ],
            }
            zones.append(new_zone)

        request_json: dict[str, Any] = {
            "away_temp": schedule.away_temp,
            "hg_temp": schedule.hg_temp,
            "timetable": timetable_entries,
            "zones": zones,
        }

        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=SYNCHOMESCHEDULE_ENDPOINT,
            params={
                "params": {
                    "home_id": self.entity_id,
                    "schedule_id": schedule.entity_id,
                    "name": "Default",
                },
                "json": request_json,
            },
        )

        return (await resp.json()).get("status") == "ok"


def is_valid_state(data: dict[str, Any]) -> bool:
    """Check set state data."""
    return data is not None


def is_valid_schedule(schedule: Schedule) -> bool:
    """Check schedule."""
    return (
        isinstance(schedule, Schedule)
        and hasattr(schedule, "entity_id")
        and schedule.entity_id != ""
    )


def get_temperature_control_mode(
    temperature_control_mode: str | None,
) -> TemperatureControlMode | None:
    """Return temperature control mode.

    Unknown values degrade to None with a warning rather than raising, so a
    single unrecognized mode never aborts topology parsing for the whole
    account.
    """
    if not temperature_control_mode:
        return None
    try:
        return TemperatureControlMode(temperature_control_mode)
    except ValueError:
        LOG.warning(
            "%s temperature control mode is unknown",
            temperature_control_mode,
        )
        return None



"""Module to represent VELUX ACTIVE modules."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pyatmo.modules.module import EntityBase, Module, Shutter, WifiMixin

if TYPE_CHECKING:
    from pyatmo.const import RawData
    from pyatmo.home import Home
    from pyatmo.modules.module import ModuleT


class VeluxGatewayMixin(EntityBase):
    """Mixin for VELUX gateway specific data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize VELUX gateway data."""
        super().__init__(home, module)
        self.subtype: str | None = None
        self.firmware_revision_netatmo: int | None = None
        self.firmware_revision_thirdparty: int | None = None
        self.hardware_version: int | None = None
        self.last_seen: int | None = None
        self.last_reset_type: str | None = None
        self.busy: bool | None = None
        self.calibrating: bool | None = None
        self.is_raining: bool | None = None
        self.locked: bool | None = None
        self.locking: bool | None = None
        self.pairing: bool | None = None
        self.pincode_enabled: bool | None = None
        self.secure: bool | None = None


class VeluxOpenerMixin(EntityBase):
    """Mixin for VELUX cover specific data."""

    def __init__(self, home: Home, module: ModuleT) -> None:
        """Initialize VELUX cover data."""
        super().__init__(home, module)
        self.last_seen: int | None = None
        self.manufacturer: str | None = None
        self.mode: str | None = None
        self.silent: bool | None = None
        self.velux_type: str | None = None


class NXG(VeluxGatewayMixin, WifiMixin, Module):
    """Class to represent a VELUX ACTIVE gateway."""

    async def update(self, raw_data: RawData) -> None:
        """Update gateway state without cascading gateway data into bridged covers."""

        self.update_topology(raw_data)
        self.update_features()


class NXO(VeluxOpenerMixin, Shutter):
    """Class to represent a VELUX ACTIVE opener / cover."""



"""Module to represent somfy modules."""

from __future__ import annotations

import logging

from pyatmo.modules.module import RfMixin, Shutter

LOG: logging.Logger = logging.getLogger(__name__)


class TPSRS(RfMixin, Shutter):
    """Class to represent a somfy TPSRS."""



"""Define tests for climate module."""

import json
import logging
from unittest.mock import AsyncMock, MagicMock, patch

import anyio
import pytest

from pyatmo import DeviceType, NoScheduleError, const
from pyatmo.modules import NATherm1
from pyatmo.modules.device_types import (
    BoilerControl,
    BoilerError,
    DeviceCategory,
    DhwControl,
)
from pyatmo.modules.module import BoilerMixin, OpenThermMixin
from pyatmo.modules.netatmo import OTH
from pyatmo.room import (
    climate_setpoint_mode_to_pilot_wire,
    pilot_wire_to_climate_setpoint_mode,
)
from tests.common import MockResponse, fake_post_request
from tests.conftest import does_not_raise


async def test_async_climate_room(async_home):
    """Test room with climate devices."""
    room_id = "2746182631"
    assert room_id in async_home.rooms

    room = async_home.rooms[room_id]
    assert room.reachable is True
    assert room.device_types == {DeviceType.NATherm1}

    module_id = "12:34:56:00:01:ae"
    assert module_id in room.modules
    assert len(room.modules) == 1


async def test_async_climate_NATherm1(async_home):
    """Test NATherm1 climate device."""
    module_id = "12:34:56:00:01:ae"
    module = async_home.modules[module_id]
    assert module.name == "Livingroom"
    assert module.device_type == DeviceType.NATherm1
    assert module.reachable is True
    assert module.boiler_status is False
    assert module.firmware_revision == 65
    assert module.battery == 75
    assert module.rf_strength == 58


async def test_async_climate_NRV(async_home):
    """Test NRV climate device."""
    module_id = "12:34:56:03:a5:54"
    module = async_home.modules[module_id]
    assert module.name == "Valve1"
    assert async_home.rooms[module.room_id].name == "Entrada"
    assert module.device_type == DeviceType.NRV
    assert module.reachable is True
    assert module.rf_strength == 51
    assert module.battery == 90
    assert module.firmware_revision == 79


async def test_async_climate_NAPlug(async_home):
    """Test NAPlug climate device."""
    module_id = "12:34:56:00:fa:d0"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NAPlug
    assert len(module.modules) == 3
    assert module.rf_strength == 107
    assert module.wifi_strength == 42
    assert module.firmware_revision == 174


async def test_async_climate_NIS(async_home):
    """Test Netatmo siren."""
    module_id = "12:34:56:00:e3:9b"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NIS
    assert module.firmware_revision == 209
    assert module.status == "no_sound"
    assert module.monitoring is False


async def test_async_climate_OTM(async_home):
    """Test OTM climate device."""
    module_id = "12:34:56:20:f5:8c"
    module = async_home.modules[module_id]
    assert module.name == "Bureau Modulate"
    assert module.device_type == DeviceType.OTM
    assert module.reachable is True
    assert module.boiler_status is False
    assert module.firmware_revision == 6
    assert module.battery == 90
    assert module.rf_strength == 64


async def test_async_climate_OTH(async_home):
    """Test OTH climate device."""
    module_id = "12:34:56:20:f5:44"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.OTH
    assert len(module.modules) == 1
    assert module.wifi_strength == 57
    assert module.firmware_revision == 22
    # coerced to enums, still string-comparable (StrEnum)
    assert module.boiler_control is BoilerControl.onoff
    assert module.boiler_error is BoilerError.water_pressure
    assert module.dhw_control is DhwControl.none
    assert module.boiler_control == "onoff"
    assert module.boiler_error == "water_pressure"
    assert module.dhw_control == "none"
    # BoilerMixin field is now parsed; defaults to None when absent from the response
    assert module.boiler_status is None


async def test_oth_boiler_enum_unknown_value():
    """Unknown enum values fall back to `unknown` instead of crashing parsing."""
    module = OTH(MagicMock(), {"id": "x", "type": "OTH"})
    module.update_topology(
        {
            "id": "x",
            "type": "OTH",
            "boiler_control": "brand_new_value",
            "boiler_error": "brand_new_error",
            "dhw_control": "brand_new_mode",
        },
    )
    assert module.boiler_control is BoilerControl.unknown
    assert module.boiler_error is BoilerError.unknown
    assert module.dhw_control is DhwControl.unknown


async def test_oth_boiler_enum_absent_no_warning(caplog):
    """Absent boiler fields stay None and do not log spurious 'unknown' warnings."""
    module = OTH(MagicMock(), {"id": "x", "type": "OTH"})
    with caplog.at_level(logging.WARNING):
        module.update_topology({"id": "x", "type": "OTH"})
    assert module.boiler_control is None
    assert module.boiler_error is None
    assert module.dhw_control is None
    assert "unknown" not in caplog.text.lower()


async def test_async_climate_OTH_mixin_composition(async_home):
    """Guard the OTH MRO: both boiler mixins run and no diagnostic field is skipped.

    Protects against reordering the OTH base classes, which could silently drop
    a mixin from the cooperative ``super().__init__`` chain.
    """
    module = async_home.modules["12:34:56:20:f5:44"]
    assert isinstance(module, BoilerMixin)
    assert isinstance(module, OpenThermMixin)
    # every diagnostic field is initialised regardless of response content
    for attr in ("boiler_control", "boiler_error", "dhw_control", "boiler_status"):
        assert hasattr(module, attr)


async def test_async_climate_BNS(async_home):
    """Test Smarther BNS climate module."""
    module_id = "10:20:30:bd:b8:1e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.BNS
    assert module.name == "Smarther"

    room = async_home.rooms[module.room_id]
    assert room.name == "Corridor"
    assert room.device_types == {
        DeviceType.BNS,
    }
    assert room.features == {"humidity", DeviceCategory.climate}


async def test_async_climate_initial_state(async_account):
    """Test initial climate state."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    home = async_account.homes[home_id]

    room_id = "2746182631"
    room = home.rooms[room_id]

    module_id = "12:34:56:00:01:ae"
    module = home.modules[module_id]
    assert room.reachable is True
    assert room.humidity is None
    assert module.name == "Livingroom"
    assert module.device_type == DeviceType.NATherm1
    assert module.reachable is True
    assert module.boiler_status is False
    assert module.battery == 75
    assert isinstance(module, NATherm1)


async def test_async_climate_last_seen(async_account):
    """last_seen is parsed for non-VELUX modules."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    home = async_account.homes[home_id]

    module = home.modules["12:34:56:00:01:ae"]
    assert module.last_seen == 1776675797
    # availability metadata, not a measurement feature
    assert "last_seen" not in module.features


async def test_async_climate_disconnected_state(async_account):
    """Test climate state when disconnected."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    home = async_account.homes[home_id]
    room_id = "2746182631"
    room = home.rooms[room_id]
    module_id = "12:34:56:00:01:ae"
    module = home.modules[module_id]

    async with await anyio.open_file(
        "fixtures/home_status_error_disconnected.json",
        encoding="utf-8",
    ) as json_file:
        content = await json_file.read()
        home_status_fixture = json.loads(content)
    mock_home_status_resp = MockResponse(home_status_fixture, 200)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=mock_home_status_resp),
    ) as mock_request:
        await async_account.async_update_status(home_id)
        mock_request.assert_called()

    assert room.reachable is None
    assert module.reachable is False


async def test_async_climate_reconnected_state(async_account):
    """Test climate state after reconnection."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_status(home_id)
    home = async_account.homes[home_id]
    room_id = "2746182631"
    room = home.rooms[room_id]
    module_id = "12:34:56:00:01:ae"
    module = home.modules[module_id]

    async with await anyio.open_file(
        "fixtures/home_status_simple.json",
        encoding="utf-8",
    ) as json_file:
        home_status_fixture = json.loads(await json_file.read())
    mock_home_status_resp = MockResponse(home_status_fixture, 200)

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=mock_home_status_resp),
    ) as mock_request:
        await async_account.async_update_status(home_id)
        mock_request.assert_called()

    assert room.reachable is True
    assert module.reachable is True
    assert module.battery == 75
    assert module.rf_strength == 58


@pytest.mark.parametrize(
    ("t_sched_id", "expected"),
    [
        ("591b54a2764ff4d50d8b5795", does_not_raise()),
        (
            "123456789abcdefg12345678",
            pytest.raises(NoScheduleError),
        ),
    ],
)
async def test_async_climate_switch_schedule(
    async_home,
    t_sched_id,
    expected,
):
    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            AsyncMock(return_value=MockResponse(response, 200)),
        ),
        expected,
    ):
        await async_home.async_switch_schedule(
            schedule_id=t_sched_id,
        )


@pytest.mark.parametrize(
    ("temp", "end_time"),
    [
        (
            14,
            None,
        ),
        (
            14,
            1559162650,
        ),
        (
            None,
            None,
        ),
        (
            None,
            1559162650,
        ),
    ],
)
async def test_async_climate_room_therm_set(
    async_home,
    temp,
    end_time,
):
    room_id = "2746182631"
    mode = "home"

    expected_params = {
        "home_id": "91763b24c43d3e344f424e8b",
        "room_id": room_id,
        "mode": mode,
    }
    if temp:
        expected_params["temp"] = str(temp)
    if end_time:
        expected_params["endtime"] = str(end_time)

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_post:
        room = async_home.rooms[room_id]

        await room.async_therm_set(
            mode=mode,
            temp=temp,
            end_time=end_time,
        )
        mock_post.assert_awaited_once_with(
            endpoint="api/setroomthermpoint",
            params=expected_params,
        )


@pytest.mark.parametrize(
    ("mode", "end_time", "schedule_id", "json_fixture", "expected", "exception"),
    [
        (
            "away",
            None,
            None,
            "status_ok.json",
            True,
            does_not_raise(),
        ),
        (
            "away",
            1559162650,
            None,
            "status_ok.json",
            True,
            does_not_raise(),
        ),
        (
            "schedule",
            None,
            "591b54a2764ff4d50d8b5795",
            "status_ok.json",
            True,
            does_not_raise(),
        ),
        (
            "schedule",
            1559162650,
            "591b54a2764ff4d50d8b5795",
            "status_ok.json",
            True,
            does_not_raise(),
        ),
        (
            None,
            None,
            None,
            "home_status_error_mode_is_missing.json",
            False,
            pytest.raises(NoScheduleError),
        ),
        (
            "away",
            1559162650,
            0000000,
            "status_ok.json",
            True,
            pytest.raises(NoScheduleError),
        ),
        (
            "schedule",
            None,
            "blahblahblah",
            "home_status_error_invalid_schedule_id.json",
            False,
            pytest.raises(NoScheduleError),
        ),
    ],
)
async def test_async_climate_set_thermmode(
    async_home,
    mode,
    end_time,
    schedule_id,
    json_fixture,
    expected,
    exception,
):
    async with await anyio.open_file(
        f"fixtures/{json_fixture}",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            AsyncMock(return_value=MockResponse(response, 200)),
        ),
        exception,
    ):
        resp = await async_home.async_set_thermmode(
            mode=mode,
            end_time=end_time,
            schedule_id=schedule_id,
        )
        assert expected is resp


async def test_async_climate_empty_home(async_account):
    """Test climate setup with empty home."""
    home_id = "91763b24c43d3e344f424e8c"

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        fake_post_request,
    ):
        await async_account.async_update_status(home_id)

    assert home_id in async_account.homes

    home = async_account.homes[home_id]
    assert len(home.rooms) == 0


async def test_power_wire(async_home_multi):
    """Test room with climate devices."""
    room_id = "3707962039"
    assert room_id in async_home_multi.rooms

    room = async_home_multi.rooms[room_id]

    assert room.climate_type == DeviceType.NLC
    assert DeviceType.NLC in room.device_types
    assert room.support_pilot_wire is True


@pytest.mark.parametrize(
    ("mode", "expected"),
    [
        (const.MANUAL, const.PILOT_WIRE_COMFORT),
        (const.MAX, const.PILOT_WIRE_COMFORT),
        (const.OFF, const.PILOT_WIRE_FROST_GUARD),
        (const.HOME, const.PILOT_WIRE_COMFORT),
        (const.FROSTGUARD, const.PILOT_WIRE_FROST_GUARD),
        (const.SCHEDULE, const.PILOT_WIRE_COMFORT),
        (const.AWAY, const.PILOT_WIRE_AWAY),
        # unknown mode falls back to frost guard
        ("nonsense", const.PILOT_WIRE_FROST_GUARD),
    ],
)
def test_climate_setpoint_mode_to_pilot_wire(mode, expected):
    """Test climate setpoint mode -> pilot wire preset mapping (frost-guard default)."""
    assert climate_setpoint_mode_to_pilot_wire(mode) == expected


@pytest.mark.parametrize(
    ("pilot_wire", "expected"),
    [
        (const.PILOT_WIRE_COMFORT, const.MANUAL),
        (const.PILOT_WIRE_AWAY, const.MANUAL),
        (const.PILOT_WIRE_FROST_GUARD, const.FROSTGUARD),
        (const.PILOT_WIRE_STAND_BY, const.FROSTGUARD),
        (const.PILOT_WIRE_COMFORT_1, const.HOME),
        (const.PILOT_WIRE_COMFORT_2, const.HOME),
        # unknown pilot wire preset falls back to frost guard
        ("nonsense", const.FROSTGUARD),
    ],
)
def test_pilot_wire_to_climate_setpoint_mode(pilot_wire, expected):
    """Test pilot wire preset -> canonical NLC setpoint mode mapping (frost-guard default)."""
    assert pilot_wire_to_climate_setpoint_mode(pilot_wire) == expected



"""Define tests for energy module."""

import datetime as dt
import json
from unittest.mock import patch

import anyio
import pytest
import time_machine

from pyatmo import ApiHomeReachabilityError, DeviceType
from pyatmo.modules.module import EnergyHistoryMixin, MeasureInterval
from tests.common import MockResponse


async def test_async_energy_NLPC(async_home):
    """Test Legrand / BTicino connected energy meter module."""
    module_id = "12:34:56:00:00:a1:4c:da"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NLPC
    assert module.power == 476


@time_machine.travel(dt.datetime(2022, 2, 12, 7, 59, 49))
async def test_historical_data_retrieval(async_account):
    """Test retrieval of historical measurements."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_events(home_id=home_id)
    home = async_account.homes[home_id]

    module_id = "12:34:56:00:00:a1:4c:da"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.device_type == DeviceType.NLPC

    await async_account.async_update_measures(home_id=home_id, module_id=module_id)
    # changed the reference here as start and stop data was not calculated in the spirit of the netatmo api where their time data is in the fact representing the "middle" of the range and not the begining
    assert module.historical_data[0] == {
        "Wh": 197,
        "duration": 60,
        "endTime": "2022-02-05T08:59:49Z",
        "endTimeUnix": 1644051589,
        "energyMode": ["basic"],
        "WhPerModes": [197],
        "startTime": "2022-02-05T07:59:50Z",
        "startTimeUnix": 1644047989,
    }
    assert module.historical_data[-1] == {
        "Wh": 259,
        "duration": 60,
        "endTime": "2022-02-12T07:59:49Z",
        "endTimeUnix": 1644652789,
        "energyMode": ["basic"],
        "WhPerModes": [259],
        "startTime": "2022-02-12T06:59:50Z",
        "startTimeUnix": 1644649189,
    }
    assert len(module.historical_data) == 168


@time_machine.travel(dt.datetime(2024, 7, 24, 22, 00, 10))
async def test_historical_data_retrieval_multi(async_account_multi):
    """Test retrieval of historical measurements."""
    home_id = "aaaaaaaaaaabbbbbbbbbbccc"

    home = async_account_multi.homes[home_id]

    module_id = "98:76:54:32:10:00:00:69"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.device_type == DeviceType.NLPC

    strt = int(dt.datetime.fromisoformat("2024-07-24 00:00:00").timestamp())
    end_time = int(dt.datetime.fromisoformat("2024-07-24 22:27:00").timestamp())

    await async_account_multi.async_update_measures(
        home_id=home_id,
        module_id=module_id,
        interval=MeasureInterval.HALF_HOUR,
        start_time=strt,
        end_time=end_time,
    )
    assert isinstance(module, EnergyHistoryMixin)

    assert module.historical_data is not None

    assert module.historical_data[0] == {
        "Wh": 20,
        "duration": 30,
        "endTime": "2024-07-23T22:30:00Z",
        "endTimeUnix": 1721773800,
        "energyMode": ["basic"],
        "WhPerModes": [20],
        "startTime": "2024-07-23T22:00:01Z",
        "startTimeUnix": 1721772000,
    }
    assert module.historical_data[17] == {
        "Wh": 710,
        "WhPerModes": [498, 212],
        "duration": 30,
        "endTime": "2024-07-24T07:00:00Z",
        "endTimeUnix": 1721804400,
        "energyMode": ["basic", "peak"],
        "startTime": "2024-07-24T06:30:01Z",
        "startTimeUnix": 1721802600,
    }

    assert module.historical_data[-1] == {
        "Wh": 16,
        "WhPerModes": [16],
        "duration": 30,
        "endTime": "2024-07-24T17:30:00Z",
        "endTimeUnix": 1721842200,
        "energyMode": ["peak"],
        "startTime": "2024-07-24T17:00:01Z",
        "startTimeUnix": 1721840400,
    }

    assert len(module.historical_data) == 39
    assert module.sum_energy_elec == 17547
    assert module.sum_energy_elec_off_peak == 4290
    assert module.sum_energy_elec_peak == 10177


@patch("pyatmo.auth.AbstractAsyncAuth.async_post_api_request")
async def test_disconnected_main_bridge(mock_home_status, async_account_multi):
    """Test retrieval of historical measurements."""
    home_id = "aaaaaaaaaaabbbbbbbbbbccc"

    async with await anyio.open_file(
        "fixtures/home_multi_status_error_disconnected.json",
        encoding="utf-8",
    ) as json_file:
        content = await json_file.read()
        home_status_fixture = json.loads(content)
    mock_home_status_resp = MockResponse(home_status_fixture, 200)
    mock_home_status.return_value = mock_home_status_resp

    with pytest.raises(ApiHomeReachabilityError):
        await async_account_multi.async_update_status(home_id)



#!/usr/bin/env python3
"""Release helper for pyatmo.

Finalizes ``CHANGELOG.md`` for a release: renames the ``[unreleased]`` section to
the new version with a date, resets a fresh empty ``[unreleased]`` scaffold,
maintains the bottom compare links, and can extract release notes for a version.

Version numbers come from git tags (``setuptools_scm``); this script only decides
the next number from a ``--bump`` type and rewrites the changelog. It never tags
or pushes -- that is done by the CI workflow (or a maintainer for a local run).

Usage:
    release.py --bump {patch,minor,major} [--dry-run]
    release.py --notes X.Y.Z
"""

from __future__ import annotations

import argparse
import datetime
import difflib
import os
from pathlib import Path
import re
import subprocess
import sys

REPO_DEFAULT = "jabesq-org/pyatmo"
CHANGELOG = Path(__file__).resolve().parents[1] / "CHANGELOG.md"

_BUMP_PARTS = ("patch", "minor", "major")
_SECTION_RE = re.compile(r"^## \[", re.MULTILINE)
_SUBSECTION_RE = re.compile(r"^### .+$", re.MULTILINE)
_VERSION_RE = re.compile(r"^\d+\.\d+\.\d+$")


def _line_end(text: str, pos: int) -> int:
    """Return the offset of the newline at/after ``pos``, or end of text."""
    nl = text.find("\n", pos)
    return len(text) if nl == -1 else nl


class ReleaseError(Exception):
    """Raised when a release cannot proceed (e.g. empty changelog)."""


def bump_version(current: str, part: str) -> str:
    """Return the next version after bumping ``part`` of ``current``.

    ``current`` may carry a leading ``v``. ``part`` is patch/minor/major.
    """
    if part not in _BUMP_PARTS:
        msg = f"unknown bump part {part!r}; expected one of {_BUMP_PARTS}"
        raise ValueError(msg)

    stripped = current.removeprefix("v")
    if not _VERSION_RE.match(stripped):
        msg = f"expected an X.Y.Z version tag, got {current!r}"
        raise ReleaseError(msg)
    major, minor, patch = (int(x) for x in stripped.split("."))
    if part == "major":
        return f"{major + 1}.0.0"
    if part == "minor":
        return f"{major}.{minor + 1}.0"
    return f"{major}.{minor}.{patch + 1}"


def _split_sections(text: str) -> list[tuple[str, int, int]]:
    """Return (header_line, start, end) for each ``## [...]`` section.

    ``start`` is the offset of the ``##`` header, ``end`` the offset of the next
    header (or the start of the link section / EOF).
    """
    matches = list(_SECTION_RE.finditer(text))
    sections: list[tuple[str, int, int]] = []
    for i, m in enumerate(matches):
        start = m.start()
        end = matches[i + 1].start() if i + 1 < len(matches) else len(text)
        header = text[start : _line_end(text, start)]
        sections.append((header, start, end))
    return sections


def _section_body(text: str, label: str) -> str | None:
    """Return the raw body (text after the header line) of section ``label``."""
    for header, start, end in _split_sections(text):
        if header.strip() == f"## [{label}]" or header.strip().startswith(
            f"## [{label}] "
        ):
            line_end = _line_end(text, start)
            header_end = min(line_end + 1, len(text))
            return text[header_end:end]
    return None


def has_unreleased_entries(text: str) -> bool:
    """Return whether ``[unreleased]`` contains at least one real bullet."""
    body = _section_body(text, "unreleased")
    if body is None:
        return False
    return any(line.startswith("- ") and line[2:].strip() for line in body.splitlines())


def strip_empty_subsections(body: str) -> str:
    """Drop ``### `` subsections whose only bullet is a ``-`` placeholder/blank."""
    matches = list(_SUBSECTION_RE.finditer(body))
    if not matches:
        return body
    kept = [body[: matches[0].start()]]
    for i, m in enumerate(matches):
        start = m.start()
        end = matches[i + 1].start() if i + 1 < len(matches) else len(body)
        chunk = body[start:end]
        content = chunk[chunk.index("\n") + 1 :] if "\n" in chunk else ""
        if any(
            line.startswith("- ") and line[2:].strip() for line in content.splitlines()
        ):
            kept.append(chunk)
    return "".join(kept)


def extract_notes(text: str, version: str) -> str:
    """Return the changelog body for ``version`` (without the header), trimmed."""
    body = _section_body(text, version)
    if body is None:
        msg = f"no changelog section found for version {version}"
        raise ReleaseError(msg)
    return body.strip()


def _latest_tag() -> str:
    """Return the latest ``v*`` git tag by version order."""
    out = subprocess.run(
        ["git", "tag", "--list", "v*", "--sort=-v:refname"],
        check=True,
        capture_output=True,
        text=True,
    ).stdout
    tags = [line.strip() for line in out.splitlines() if line.strip()]
    if not tags:
        msg = "no v* git tags found"
        raise ReleaseError(msg)
    return tags[0]


def finalize_changelog(
    text: str,
    version: str,
    date: str,
    repo: str = REPO_DEFAULT,
) -> str:
    """Return ``text`` finalized for ``version``.

    Renames ``## [unreleased]`` to ``## [version] - date``, inserts a fresh empty
    ``## [unreleased]`` on top, and updates the bottom compare links.
    """
    if not has_unreleased_entries(text):
        msg = "the [unreleased] section has no entries; nothing to release"
        raise ReleaseError(msg)

    prev = _unreleased_prev_version(text)

    # Locate the unreleased section and strip its empty subsections.
    start = end = header_end = None
    for header, sec_start, sec_end in _split_sections(text):
        if header.strip() == "## [unreleased]":
            start, end = sec_start, sec_end
            header_end = min(_line_end(text, sec_start) + 1, len(text))
            break
    if start is None:
        msg = "could not find the [unreleased] section"
        raise ReleaseError(msg)

    cleaned = strip_empty_subsections(text[header_end:end])

    # Rename the section to the new version and add a fresh empty unreleased.
    replacement = f"## [unreleased]\n\n## [{version}] - {date}\n{cleaned}"
    new_text = text[:start] + replacement + text[end:]

    # Maintain the bottom compare links.
    base = f"https://github.com/{repo}/compare"
    return re.sub(
        r"^\[unreleased\]: .*$",
        f"[unreleased]: {base}/v{version}...HEAD\n"
        f"[{version}]: {base}/v{prev}...v{version}",
        new_text,
        count=1,
        flags=re.MULTILINE,
    )


def _unreleased_prev_version(text: str) -> str:
    """Return the version the ``[unreleased]`` compare link currently points at."""
    m = re.search(
        r"^\[unreleased\]: .*/compare/v([0-9.]+)\.\.\.HEAD", text, re.MULTILINE
    )
    if not m:
        msg = "could not find the [unreleased] compare link"
        raise ReleaseError(msg)
    return m.group(1)


def _emit_output(version: str, notes: str) -> None:
    """Write step outputs when running inside GitHub Actions."""
    gh_output = os.environ.get("GITHUB_OUTPUT")
    if not gh_output:
        return
    with Path(gh_output).open("a", encoding="utf-8") as fh:
        fh.write(f"version={version}\n")
        delimiter = "NOTES_EOF"
        fh.write(f"notes<<{delimiter}\n{notes}\n{delimiter}\n")


def _cmd_bump(args: argparse.Namespace) -> int:
    text = CHANGELOG.read_text(encoding="utf-8")
    current = _latest_tag()
    version = bump_version(current, args.bump)
    date = datetime.datetime.now(tz=datetime.UTC).strftime("%Y-%m-%d")

    try:
        new_text = finalize_changelog(text, version, date)
    except ReleaseError as err:
        print(f"error: {err}", file=sys.stderr)
        return 1

    notes = extract_notes(new_text, version)

    if args.dry_run:
        print(f"Next version: {version} (from {current})")
        print("\n--- release notes ---")
        print(notes)
        print("\n--- changelog changes ---")
        _print_diff(text, new_text)
        return 0

    CHANGELOG.write_text(new_text, encoding="utf-8")
    _emit_output(version, notes)
    print(f"Finalized CHANGELOG.md for {version}")
    return 0


def _cmd_notes(args: argparse.Namespace) -> int:
    text = CHANGELOG.read_text(encoding="utf-8")
    print(extract_notes(text, args.notes))
    return 0


def _print_diff(before: str, after: str) -> None:
    diff = difflib.unified_diff(
        before.splitlines(keepends=True),
        after.splitlines(keepends=True),
        fromfile="CHANGELOG.md",
        tofile="CHANGELOG.md (new)",
    )
    sys.stdout.writelines(diff)


def main(argv: list[str] | None = None) -> int:
    """Parse arguments and dispatch to the requested subcommand."""
    parser = argparse.ArgumentParser(description=__doc__)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--bump", choices=_BUMP_PARTS, help="version part to bump")
    group.add_argument("--notes", metavar="X.Y.Z", help="print notes for a version")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="preview only; write nothing and touch no git state",
    )
    args = parser.parse_args(argv)

    if args.notes:
        return _cmd_notes(args)
    return _cmd_bump(args)


if __name__ == "__main__":
    raise SystemExit(main())



"""Support for a Netatmo account."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast
from uuid import uuid4

from pyatmo import modules
from pyatmo.const import (
    GETEVENTS_ENDPOINT,
    GETHOMECOACHDATA_ENDPOINT,
    GETHOMESDATA_ENDPOINT,
    GETHOMESTATUS_ENDPOINT,
    GETPUBLIC_DATA_ENDPOINT,
    GETSTATIONDATA_ENDPOINT,
    HOME,
    SETSTATE_ENDPOINT,
    RawData,
)
from pyatmo.enums import PressureUnit, UnitSystem, WindUnit
from pyatmo.helpers import extract_raw_data
from pyatmo.home import Home
from pyatmo.modules.module import Energy, MeasureInterval, Module

if TYPE_CHECKING:
    from aiohttp import ClientResponse

    from pyatmo.auth import AbstractAsyncAuth

LOG: logging.Logger = logging.getLogger(__name__)


class AsyncAccount:
    """Async class of a Netatmo account."""

    def __init__(
        self,
        auth: AbstractAsyncAuth,
        favorite_stations: bool = True,
    ) -> None:
        """Initialize the Netatmo account."""

        self.auth: AbstractAsyncAuth = auth
        self.user: str | None = None
        self.user_country: str | None = None
        self.pending_user_consent: bool | None = None
        self.unit_system: UnitSystem | None = None
        self.unit_wind: WindUnit | None = None
        self.unit_pressure: PressureUnit | None = None
        self.all_homes_id: dict[str, str] = {}
        self.homes: dict[str, Home] = {}
        self.raw_data: RawData = {}
        self.favorite_stations: bool = favorite_stations
        self.public_weather_areas: dict[str, modules.PublicWeatherArea] = {}
        self.modules: dict[str, Module] = {}

    def __repr__(self) -> str:
        """Return the representation."""

        return (
            f"{self.__class__.__name__}(user={self.user}, home_ids={self.homes.keys()}"
        )

    def process_topology(self, disabled_homes_ids: list[str] | None = None) -> None:
        """Process topology information from /homesdata."""

        if disabled_homes_ids is None:
            disabled_homes_ids = []

        for home in self.raw_data["homes"]:
            home_id: str = home.get("id", "Unknown")
            home_name: str = home.get("name", "Unknown")
            self.all_homes_id[home_id] = home_name

            if home_id in disabled_homes_ids:
                if home_id in self.homes:
                    del self.homes[home_id]
                continue

            if home_id in self.homes:
                self.homes[home_id].update_topology(home)
            else:
                self.homes[home_id] = Home(self.auth, raw_data=home)

    async def async_update_topology(
        self,
        disabled_homes_ids: list[str] | None = None,
    ) -> None:
        """Retrieve topology data from /homesdata."""

        resp = await self.auth.async_post_api_request(
            endpoint=GETHOMESDATA_ENDPOINT,
        )
        body = await resp.json()
        self.raw_data = extract_raw_data(body, "homes")

        # Read the user block straight from the response body; keep it out of
        # raw_data so consumers that serialize raw_data (e.g. Home Assistant
        # diagnostics) do not leak the user's email/id.
        user = body.get("body", {}).get("user", {})
        self.user = user.get("email")
        self.user_country = user.get("country")
        self.pending_user_consent = user.get("pending_user_consent")
        unit_system = user.get("unit_system")
        unit_wind = user.get("unit_wind")
        unit_pressure = user.get("unit_pressure")
        self.unit_system = None if unit_system is None else UnitSystem(unit_system)
        self.unit_wind = None if unit_wind is None else WindUnit(unit_wind)
        self.unit_pressure = (
            None if unit_pressure is None else PressureUnit(unit_pressure)
        )

        self.process_topology(disabled_homes_ids=disabled_homes_ids)

    async def async_update_status(self, home_id: str) -> None:
        """Retrieve status data from /homestatus."""
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=GETHOMESTATUS_ENDPOINT,
            params={"home_id": home_id},
        )
        raw_data: RawData = extract_raw_data(await resp.json(), HOME)
        await self.homes[home_id].update(raw_data, do_raise_for_reachability_error=True)

    async def async_update_events(self, home_id: str) -> None:
        """Retrieve events from /getevents."""
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=GETEVENTS_ENDPOINT,
            params={"home_id": home_id},
        )
        raw_data: RawData = extract_raw_data(await resp.json(), HOME)
        await self.homes[home_id].update(raw_data)

    async def async_update_weather_stations(self) -> None:
        """Retrieve status data from /getstationsdata."""
        params: dict[str, str] = {
            "get_favorites": ("true" if self.favorite_stations else "false")
        }
        await self._async_update_data(
            GETSTATIONDATA_ENDPOINT,
            params=params,
        )

    async def async_update_air_care(self) -> None:
        """Retrieve status data from /gethomecoachsdata."""
        await self._async_update_data(GETHOMECOACHDATA_ENDPOINT)

    async def async_update_measures(
        self,
        home_id: str,
        module_id: str,
        start_time: int | None = None,
        end_time: int | None = None,
        interval: MeasureInterval = MeasureInterval.HOUR,
        days: int = 7,
    ) -> None:
        """Retrieve measures data from /getmeasure."""

        module: Module = self.homes[home_id].modules[module_id]
        if module.has_feature("historical_data"):
            module = cast("Energy", module)
            await module.async_update_measures(
                start_time=start_time,
                end_time=end_time,
                interval=interval,
                days=days,
            )

    def register_public_weather_area(
        self,
        lat_ne: str,
        lon_ne: str,
        lat_sw: str,
        lon_sw: str,
        required_data_type: str | None = None,
        filtering: bool = False,
        *,
        area_id: str | None = None,
    ) -> str:
        """Register public weather area to monitor."""

        if area_id is None:
            area_id = str(uuid4())

        self.public_weather_areas[area_id] = modules.PublicWeatherArea(
            lat_ne,
            lon_ne,
            lat_sw,
            lon_sw,
            required_data_type,
            filtering,
        )
        return area_id

    async def async_update_public_weather(self, area_id: str) -> None:
        """Retrieve status data from /getpublicdata."""
        params: dict[str, str] = {
            "lat_ne": self.public_weather_areas[area_id].location.lat_ne,
            "lon_ne": self.public_weather_areas[area_id].location.lon_ne,
            "lat_sw": self.public_weather_areas[area_id].location.lat_sw,
            "lon_sw": self.public_weather_areas[area_id].location.lon_sw,
            "filtering": (
                "true" if self.public_weather_areas[area_id].filtering else "false"
            ),
        }
        await self._async_update_data(
            GETPUBLIC_DATA_ENDPOINT,
            tag="body",
            params=params,
            area_id=area_id,
        )

    async def _async_update_data(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
        tag: str = "devices",
        area_id: str | None = None,
    ) -> None:
        """Retrieve status data from <endpoint>."""
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=endpoint, params=params
        )
        raw_data: RawData = extract_raw_data(await resp.json(), tag)
        await self.update_devices(raw_data, area_id)

    async def async_set_state(self, home_id: str, data: dict[str, Any]) -> None:
        """Modify device state by passing JSON specific to the device."""
        LOG.debug("Setting state: %s", data)

        post_params: dict[str, Any] = {
            "json": {
                HOME: {
                    "id": home_id,
                    **data,
                },
            },
        }
        resp: ClientResponse = await self.auth.async_post_api_request(
            endpoint=SETSTATE_ENDPOINT,
            params=post_params,
        )
        LOG.debug("Response: %s", resp)

    async def update_devices(
        self,
        raw_data: RawData,
        area_id: str | None = None,
    ) -> None:
        """Update device states."""
        for device_data in raw_data.get("devices", {}):
            if home_id := device_data.get(
                "home_id",
                self.find_home_of_device(device_data),
            ):
                if home_id not in self.homes:
                    modules_data: list[dict[str, Any]] = []
                    for module_data in device_data.get("modules", []):
                        module_data["home_id"] = home_id
                        module_data["id"] = module_data["_id"]
                        module_data["name"] = module_data.get("module_name")
                        modules_data.append(normalize_weather_attributes(module_data))
                    modules_data.append(normalize_weather_attributes(device_data))

                    self.homes[home_id] = Home(
                        self.auth,
                        raw_data={
                            "id": home_id,
                            "name": device_data.get("home_name", "Unknown"),
                            "modules": modules_data,
                        },
                    )
                await self.homes[home_id].update(
                    {HOME: {"modules": [normalize_weather_attributes(device_data)]}},
                )
            else:
                LOG.debug("No home %s (%s) found.", home_id, home_id)

            for module_data in device_data.get("modules", []):
                module_data["home_id"] = home_id
                await self.update_devices({"devices": [module_data]})

            if (
                device_data["type"] == "NHC"
                or self.find_home_of_device(device_data) is None
            ):
                device_data["name"] = device_data.get(
                    "station_name",
                    device_data.get("module_name", "Unknown"),
                )
                device_data = normalize_weather_attributes(device_data)
                if device_data["id"] not in self.modules:
                    module_class: Any = getattr(
                        modules,
                        device_data["type"],
                        None,
                    )
                    if module_class is None:
                        LOG.info("Unknown device type %s", device_data["type"])
                        module_class = modules.NLunknown
                    self.modules[device_data["id"]] = module_class(
                        home=self,
                        module=device_data,
                    )
                await self.modules[device_data["id"]].update(device_data)

                if device_data.get("modules", []):
                    self.modules[device_data["id"]].modules = [
                        module["_id"] for module in device_data["modules"]
                    ]

        if area_id is not None:
            self.public_weather_areas[area_id].update(raw_data)

    def find_home_of_device(self, device_data: dict[str, Any]) -> str | None:
        """Find home_id of device."""
        return next(
            (
                home_id
                for home_id, home in self.homes.items()
                if device_data["_id"] in home.modules
            ),
            None,
        )


ATTRIBUTES_TO_FIX: dict[str, str] = {
    "_id": "id",
    "firmware": "firmware_revision",
    "wifi_status": "wifi_strength",
    "rf_status": "rf_strength",
    "Temperature": "temperature",
    "Humidity": "humidity",
    "Pressure": "pressure",
    "CO2": "co2",
    "AbsolutePressure": "absolute_pressure",
    "Noise": "noise",
    "Rain": "rain",
    "WindStrength": "wind_strength",
    "WindAngle": "wind_angle",
    "GustStrength": "gust_strength",
    "GustAngle": "gust_angle",
}


def normalize_weather_attributes(raw_data: RawData) -> dict[str, Any]:
    """Normalize weather attributes."""
    result: dict[str, Any] = {}
    for attribute, value in raw_data.items():
        if attribute == "dashboard_data":
            result.update(**normalize_weather_attributes(value))
        else:
            result[ATTRIBUTES_TO_FIX.get(attribute, attribute)] = value
    return result



"""Module to represent a Netatmo event."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyatmo.const import RawData

EVENT_ATTRIBUTES_MAP: dict[str, str] = {
    "id": "entity_id",
    "type": "event_type",
    "time": "event_time",
}


class EventTypes(Enum):
    """Event types."""

    movement = "movement"
    person = "person"
    person_away = "person_away"
    connection = "connection"
    disconnection = "disconnection"
    new_module = "new_module"
    module_connect = "module_connect"
    module_disconnect = "module_disconnect"
    module_low_battery = "module_low_battery"
    module_end_update = "module_end_update"
    on = "on"
    off = "off"
    sd = "sd"
    alim = "alim"
    boot = "boot"
    outdoor = "outdoor"
    daily_summary = "daily_summary"
    tag_big_move = "tag_big_move"
    tag_small_move = "tag_small_move"
    tag_uninstalled = "tag_uninstalled"
    tag_open = "tag_open"
    hush = "hush"
    smoke = "smoke"
    tampered = "tampered"
    wifi_status = "wifi_status"
    battery_status = "battery_status"
    detection_chamber_status = "detection_chamber_status"
    sound_test = "sound_test"
    siren_sounding = "siren_sounding"
    siren_tampered = "siren_tampered"
    incoming_call = "incoming_call"
    accepted_call = "accepted_call"
    missed_call = "missed_call"
    co_detected = "co_detected"


class VideoStatus(Enum):
    """Video states."""

    available = "available"
    deleted = "deleted"


@dataclass
class Snapshot:
    """Class to represent a Netatmo event snapshot."""

    snapshot_id: str
    version: int
    key: str
    url: str


@dataclass
class Event:
    """Class to represent a Netatmo events."""

    entity_id: str
    event_type: EventTypes
    event_time: int
    message: str | None = None
    camera_id: str | None = None
    device_id: str | None = None
    module_id: str | None = None
    person_id: str | None = None
    video_id: str | None = None
    sub_type: int | None = None
    snapshot: Snapshot | None = None
    vignette: Snapshot | None = None
    video_status: VideoStatus | None = None
    is_arrival: bool | None = None
    subevents: list[Event] | None = None

    def __init__(self, home_id: str, raw_data: RawData) -> None:
        """Initialize a Netatmo event instance."""

        self.home_id = home_id
        self._init_attributes(raw_data)

    def _init_attributes(self, raw_data: RawData) -> None:
        """Initialize attributes of the instance."""

        for attrib, value in raw_data.items():
            if attrib == "subevents":
                value = [Event(self.home_id, event) for event in value]
            setattr(self, EVENT_ATTRIBUTES_MAP.get(attrib, attrib), value)



"""Define tests for shutter module."""

import json
import logging
from unittest.mock import AsyncMock, patch

import anyio

from pyatmo import DeviceType
from tests.common import MockResponse


async def test_async_shutter_NBR(async_home):
    """Test NLP Bubendorf iDiamant roller shutter."""
    module_id = "0009999992"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NBR
    assert module.firmware_revision == 16
    assert module.current_position == 0


async def test_async_shutter_Z3V(async_home):
    """Test NLG Legrand roller shutter 3rd party."""
    module_id = "12:34:56:80:00:c3:69:3d"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.Z3V
    assert module.firmware_revision == 16
    assert module.current_position == 0


async def test_async_shutter_NBO(async_home):
    """Test NBO Bubendorf iDiamant roller shutter."""
    module_id = "0009999993"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NBO
    assert module.firmware_revision == 22
    assert module.current_position == 0


async def test_async_shutters(async_home):
    """Test basic shutter functionality."""
    room_id = "3688132631"
    assert room_id in async_home.rooms

    module_id = "0009999992"
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NBR

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    def gen_json_data(position):
        return {
            "json": {
                "home": {
                    "id": "91763b24c43d3e344f424e8b",
                    "modules": [
                        {
                            "bridge": "12:34:56:30:d5:d4",
                            "id": module_id,
                            "target_position": position,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        assert await module.async_open()
        mock_resp.assert_awaited_with(
            params=gen_json_data(100),
            endpoint="api/setstate",
        )

        assert await module.async_close()
        mock_resp.assert_awaited_with(
            params=gen_json_data(0),
            endpoint="api/setstate",
        )

        assert await module.async_stop()
        mock_resp.assert_awaited_with(
            params=gen_json_data(-1),
            endpoint="api/setstate",
        )

        assert await module.async_move_to_preferred_position()
        mock_resp.assert_awaited_with(
            params=gen_json_data(-2),
            endpoint="api/setstate",
        )

        assert await module.async_set_target_position(47)
        mock_resp.assert_awaited_with(
            params=gen_json_data(47),
            endpoint="api/setstate",
        )

        assert await module.async_set_target_position(-10)
        mock_resp.assert_awaited_with(
            params=gen_json_data(-1),
            endpoint="api/setstate",
        )

        assert await module.async_set_target_position(101)
        mock_resp.assert_awaited_with(
            params=gen_json_data(100),
            endpoint="api/setstate",
        )


async def test_async_shutter_returns_false_on_setstate_body_errors(async_home, caplog):
    """Test shutter command with setstate body errors."""
    module_id = "0009999992"
    module = async_home.modules[module_id]

    response = {
        "status": "ok",
        "body": {
            "home": {"id": "91763b24c43d3e344f424e8b"},
            "errors": [{"code": 9, "id": module_id}],
        },
    }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        with caplog.at_level(logging.WARNING, logger="pyatmo.home"):
            assert not await module.async_open()
        mock_resp.assert_awaited_with(
            params={
                "json": {
                    "home": {
                        "id": "91763b24c43d3e344f424e8b",
                        "modules": [
                            {
                                "bridge": "12:34:56:30:d5:d4",
                                "id": module_id,
                                "target_position": 100,
                            },
                        ],
                    },
                },
            },
            endpoint="api/setstate",
        )
        assert len(caplog.records) == 1
        record = caplog.records[0]
        assert record.msg == (
            "Set state response for home %s contains errors: status=%r errors=%r"
        )
        assert record.args == (
            "91763b24c43d3e344f424e8b",
            "ok",
            response["body"]["errors"],
        )



"""Module to represent Smarther modules."""

from __future__ import annotations

import logging

from pyatmo.modules.module import (
    BoilerMixin,
    CoolerMixin,
    FirmwareMixin,
    Module,
    WifiMixin,
)

LOG: logging.Logger = logging.getLogger(__name__)


class BNS(FirmwareMixin, BoilerMixin, CoolerMixin, WifiMixin, Module):
    """Smarther thermostat."""



"""Define tests for camera module."""

import json
from unittest.mock import AsyncMock, patch

import anyio
import pytest

from pyatmo import SIREN_BASE_URL, ApiError, DeviceType, WebRTCStream
from pyatmo.modules.device_types import DeviceCategory, DoorTagCategory
from tests.common import MockResponse
from tests.conftest import does_not_raise


async def test_async_doortag_NACamDoorTag(async_home):
    """NACamDoorTag exposes doortag_category, keeps device_category, no feature leak."""
    module_id = "12:34:56:00:86:99"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NACamDoorTag
    assert module.doortag_category == DoorTagCategory.window
    assert module.device_category == DeviceCategory.opening
    assert "doortag_category" not in module.features
    assert "device_category" not in module.features


async def test_async_camera_NACamera(async_home):
    """Test Netatmo indoor camera module."""
    module_id = "12:34:56:00:f1:62"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    await module.async_update_camera_urls()
    assert module.device_type == DeviceType.NACamera
    assert module.is_local
    assert module.local_url == "http://192.168.0.123/678460a0d47e5618699fb31169e2b47d"
    vpn_url = "https://prodvpn-eu-2.netatmo.net/restricted/10.255.123.45/609e27de5699fb18147ab47d06846631/MTRPn_BeWCav5RBq4U1OMDruTW4dkQ0NuMwNDAw11g,,"
    assert module.vpn_url == vpn_url
    assert module.camera_url == module.local_url
    person_id = "91827374-7e04-5298-83ad-a0cb8372dff1"
    assert person_id in module.home.persons
    person = module.home.persons[person_id]
    assert person.pseudo == "John Doe"
    assert person.out_of_sight
    assert person.last_seen == 1557071156


async def test_async_camera_NPC(async_home):
    """Test Netatmo indoor camera advance module."""
    module_id = "12:34:56:00:f1:63"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    await module.async_update_camera_urls()
    assert module.device_type == DeviceType.NPC
    assert module.is_local
    assert module.local_url is None
    vpn_url = "https://prodvpn-eu-2.netatmo.net/restricted/10.255.123.45/9fb1814609e27de5697ab47d06846631/MTRPn_BeWCav5RBq4U1OMDruTW4dkQ0NuMwNDAw11g,,"
    assert module.vpn_url == vpn_url
    assert module.camera_url == module.vpn_url
    person_id = "91827374-7e04-5298-83ad-a0cb8372dff1"
    assert person_id in module.home.persons
    person = module.home.persons[person_id]
    assert person.pseudo == "John Doe"
    assert person.out_of_sight
    assert person.last_seen == 1557071156


async def test_async_NOC(async_home):
    """Test basic outdoor camera functionality."""
    module_id = "12:34:56:10:b9:0e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NOC
    assert module.firmware_revision == 3002000
    assert module.firmware_name == "3.2.0"
    assert module.monitoring is True
    assert module.alim_status == 2
    assert module.is_local is False
    assert module.local_url is None
    vpn_url = "https://prodvpn-eu-6.netatmo.net/10.20.30.41/333333333333/444444444444,,"
    assert module.vpn_url == vpn_url
    assert module.camera_url == module.vpn_url
    assert module.floodlight == "auto"
    assert module.siren_status == "no_sound"

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    def gen_json_data(state):
        return {
            "json": {
                "home": {
                    "id": "91763b24c43d3e344f424e8b",
                    "modules": [
                        {
                            "id": module_id,
                            "floodlight": state,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        assert await module.async_floodlight_on()
        mock_resp.assert_awaited_with(
            params=gen_json_data("on"),
            endpoint="api/setstate",
        )

        assert await module.async_floodlight_off()
        mock_resp.assert_awaited_with(
            params=gen_json_data("off"),
            endpoint="api/setstate",
        )

        assert await module.async_floodlight_auto()
        mock_resp.assert_awaited_with(
            params=gen_json_data("auto"),
            endpoint="api/setstate",
        )


async def test_async_camera_monitoring(async_home):
    """Test basic camera monitoring functionality."""
    module_id = "12:34:56:10:b9:0e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NOC
    assert module.is_local is False

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    def gen_json_data(state):
        return {
            "json": {
                "home": {
                    "id": "91763b24c43d3e344f424e8b",
                    "modules": [
                        {
                            "id": module_id,
                            "monitoring": state,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        assert await module.async_monitoring_on()
        mock_resp.assert_awaited_with(
            params=gen_json_data("on"),
            endpoint="api/setstate",
        )

        assert await module.async_monitoring_off()
        mock_resp.assert_awaited_with(
            params=gen_json_data("off"),
            endpoint="api/setstate",
        )


async def test_async_camera_siren(async_home):
    """Test siren control via default public API endpoint (api.netatmo.com)."""
    module_id = "12:34:56:10:b9:0e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NOC
    assert module.siren_status == "no_sound"

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    def gen_json_data(state):
        return {
            "json": {
                "home": {
                    "id": "91763b24c43d3e344f424e8b",
                    "modules": [
                        {
                            "id": module_id,
                            "siren_status": state,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        assert await module.async_siren_on()
        mock_resp.assert_awaited_with(
            endpoint="api/setstate",
            base_url=None,
            params=gen_json_data("sound"),
        )

        assert await module.async_siren_off()
        mock_resp.assert_awaited_with(
            endpoint="api/setstate",
            base_url=None,
            params=gen_json_data("no_sound"),
        )


async def test_async_camera_siren_app_endpoint(async_home):
    """Test siren control via app.netatmo.net to bypass public API restriction.

    The public OAuth2 API (api.netatmo.com) rejects siren_status with error
    code 21. app.netatmo.net accepts the same OAuth2 token and payload.
    Callers pass base_url=SIREN_BASE_URL to opt in to this workaround.
    """
    module_id = "12:34:56:10:b9:0e"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NOC

    async with await anyio.open_file(
        "fixtures/status_ok.json",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    def gen_json_data(state):
        return {
            "json": {
                "home": {
                    "id": "91763b24c43d3e344f424e8b",
                    "modules": [
                        {
                            "id": module_id,
                            "siren_status": state,
                        },
                    ],
                },
            },
        }

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
        AsyncMock(return_value=MockResponse(response, 200)),
    ) as mock_resp:
        assert await module.async_siren_on(base_url=SIREN_BASE_URL)
        mock_resp.assert_awaited_with(
            endpoint="api/setstate",
            base_url=SIREN_BASE_URL,
            params=gen_json_data("sound"),
        )

        assert await module.async_siren_off(base_url=SIREN_BASE_URL)
        mock_resp.assert_awaited_with(
            endpoint="api/setstate",
            base_url=SIREN_BASE_URL,
            params=gen_json_data("no_sound"),
        )


async def test_async_camera_siren_missing_status(async_home):
    """Test that NOC handles missing siren_status in API payload gracefully."""
    module_id = "12:34:56:10:b9:0e"
    module = async_home.modules[module_id]

    # Simulate an API response without siren_status (e.g. older firmware)
    module.siren_status = None
    assert module.siren_status is None


@pytest.mark.parametrize(
    ("module_id", "device_type", "can_use_local_url"),
    [
        ("12:34:56:00:f1:62", DeviceType.NACamera, True),
        ("12:34:56:10:b9:0e", DeviceType.NOC, True),
        ("12:34:56:00:f1:63", DeviceType.NPC, False),
    ],
)
async def test_async_live_snapshot(
    async_home, module_id, device_type, can_use_local_url
):
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == device_type
    await module.async_update_camera_urls()
    assert module.local_url or module.vpn_url

    expected_snapshot = b"test stream image bytes"

    with patch(
        "pyatmo.auth.AbstractAsyncAuth.async_get_image",
        AsyncMock(return_value=expected_snapshot),
    ) as mock_resp:
        received_snapshot = await module.async_get_live_snapshot()

        if can_use_local_url and module.is_local:
            base_url = module.local_url
        else:
            base_url = module.vpn_url

        mock_resp.assert_awaited_with(
            base_url=base_url,
            endpoint="/live/snapshot_720.jpg",
        )

        assert received_snapshot == expected_snapshot


@pytest.mark.parametrize(
    ("response_fixture", "exception"),
    [
        ("webrtc_offer_ok.json", does_not_raise()),
        ("webrtc_offer_unreachable.json", pytest.raises(ApiError)),
    ],
)
async def test_async_webrtc_stream_start(async_home, response_fixture, exception):
    """Test starting a WebRTC stream."""
    module_id = "12:34:56:00:f1:63"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NPC

    async with await anyio.open_file(
        f"fixtures/{response_fixture}",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    session_id = "af6da83b-1fd5-46ab-bf08-8e5db3cb9725"
    sdp_offer = "sdp_test_offer"

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            AsyncMock(return_value=MockResponse(response, 200)),
        ) as mock_resp,
        exception,
    ):
        answer = await module.async_start_stream(session_id, sdp_offer)

        mock_resp.assert_awaited_with(
            params={
                "home_id": async_home.entity_id,
                "device_id": module_id,
                "session_id": session_id,
                "sdp": sdp_offer,
            },
            endpoint="api/webrtc/offer",
        )

        assert answer.stream.session_id == session_id
        assert answer.stream.tag_id == "OZzgKVlQCW0="
        assert answer.sdp == "sdp_test_answser"


@pytest.mark.parametrize(
    ("response_fixture", "exception"),
    [
        ("webrtc_terminate_ok.json", does_not_raise()),
        ("webrtc_terminate_no_session.json", pytest.raises(ApiError)),
    ],
)
async def test_async_webrtc_stream_stop(async_home, response_fixture, exception):
    """Test stopping a WebRTC stream."""
    module_id = "12:34:56:00:f1:63"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NPC

    async with await anyio.open_file(
        f"fixtures/{response_fixture}",
        encoding="utf-8",
    ) as json_file:
        response = json.loads(await json_file.read())

    session_id = "af6da83b-1fd5-46ab-bf08-8e5db3cb9725"
    tag_id = "OZzgKVlQCW0="

    with (
        patch(
            "pyatmo.auth.AbstractAsyncAuth.async_post_api_request",
            AsyncMock(return_value=MockResponse(response, 200)),
        ) as mock_resp,
        exception,
    ):
        await module.async_stop_stream(WebRTCStream(session_id, tag_id))

        mock_resp.assert_awaited_with(
            params={
                "home_id": async_home.entity_id,
                "device_id": module_id,
                "session_id": session_id,
                "tag_id": tag_id,
            },
            endpoint="api/webrtc/terminate",
        )



"""Tests for the release helper script (scripts/release.py)."""

import importlib.util
from pathlib import Path

import pytest

# Load scripts/release.py as a module (it lives outside the package).
_SCRIPT = Path(__file__).resolve().parents[1] / "scripts" / "release.py"
_spec = importlib.util.spec_from_file_location("release", _SCRIPT)
release = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(release)


REPO = "jabesq-org/pyatmo"

SAMPLE = """\
# Changelog

All notable changes to this project will be documented in this file.

## [unreleased]

### Added

- A shiny new feature (#123)

### Fixed

- A nasty bug (#124)

### Removed

-

## [9.4.0]

### Added

- Older thing

[unreleased]: https://github.com/jabesq-org/pyatmo/compare/v9.4.0...HEAD
[9.4.0]: https://github.com/jabesq-org/pyatmo/compare/v9.3.0...v9.4.0
"""

EMPTY_UNRELEASED = """\
# Changelog

## [unreleased]

## [9.4.0]

### Added

- Older thing

[unreleased]: https://github.com/jabesq-org/pyatmo/compare/v9.4.0...HEAD
"""


class TestBumpVersion:
    def test_patch(self):
        assert release.bump_version("9.4.0", "patch") == "9.4.1"

    def test_minor(self):
        assert release.bump_version("9.4.2", "minor") == "9.5.0"

    def test_major(self):
        assert release.bump_version("9.4.2", "major") == "10.0.0"

    def test_strips_leading_v(self):
        assert release.bump_version("v9.4.0", "minor") == "9.5.0"

    def test_rejects_unknown_part(self):
        with pytest.raises(ValueError, match="patch"):
            release.bump_version("9.4.0", "sideways")

    def test_rejects_two_component_tag(self):
        with pytest.raises(release.ReleaseError, match=r"9\.4"):
            release.bump_version("9.4", "patch")

    def test_rejects_non_numeric_tag(self):
        with pytest.raises(release.ReleaseError):
            release.bump_version("9.4.x", "patch")


class TestParsingRobustness:
    def test_header_as_last_line_without_newline(self):
        # A header with no trailing newline must not raise; body is empty.
        text = "# Changelog\n\n## [9.4.0]"
        assert release.extract_notes(text, "9.4.0") == ""

    def test_extract_notes_last_section_no_trailing_newline(self):
        text = "# C\n\n## [9.4.0]\n\n### Added\n\n- old"
        assert release.extract_notes(text, "9.4.0") == "### Added\n\n- old"


class TestUnreleasedGuard:
    def test_detects_entries(self):
        assert release.has_unreleased_entries(SAMPLE) is True

    def test_empty_section_has_no_entries(self):
        assert release.has_unreleased_entries(EMPTY_UNRELEASED) is False


class TestFinalizeChangelog:
    def test_renames_unreleased_with_date(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        assert "## [9.5.0] - 2026-07-20" in out

    def test_keeps_fresh_empty_unreleased_on_top(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        # unreleased header still present and appears before the new version
        assert out.index("## [unreleased]") < out.index("## [9.5.0]")
        # the released entries moved under the version, not under unreleased
        assert release.has_unreleased_entries(out) is False

    def test_moves_entries_under_new_version(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        notes = release.extract_notes(out, "9.5.0")
        assert "A shiny new feature (#123)" in notes
        assert "A nasty bug (#124)" in notes

    def test_adds_compare_link_for_new_version(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        assert (
            "[9.5.0]: https://github.com/jabesq-org/pyatmo/compare/v9.4.0...v9.5.0"
            in out
        )

    def test_repoints_unreleased_compare_link(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        assert (
            "[unreleased]: https://github.com/jabesq-org/pyatmo/compare/v9.5.0...HEAD"
            in out
        )
        assert "compare/v9.4.0...HEAD" not in out

    def test_guard_raises_on_empty_unreleased(self):
        with pytest.raises(release.ReleaseError):
            release.finalize_changelog(
                EMPTY_UNRELEASED, "9.5.0", "2026-07-20", repo=REPO
            )

    def test_is_deterministic(self):
        a = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        b = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        assert a == b


class TestStripEmptySubsections:
    def test_drops_placeholder_only_subsection(self):
        body = "\n### Added\n\n- real (#1)\n\n### Removed\n\n-\n"
        out = release.strip_empty_subsections(body)
        assert "### Added" in out
        assert "### Removed" not in out
        assert "- real (#1)" in out

    def test_keeps_subsections_with_entries(self):
        body = "\n### Added\n\n- one\n\n### Fixed\n\n- two\n"
        out = release.strip_empty_subsections(body)
        assert "### Added" in out
        assert "### Fixed" in out

    def test_drops_blank_subsection(self):
        body = "\n### Changed\n\n\n"
        assert "### Changed" not in release.strip_empty_subsections(body)


class TestFinalizeStripsEmptySections:
    def test_empty_removed_not_in_notes(self):
        out = release.finalize_changelog(SAMPLE, "9.5.0", "2026-07-20", repo=REPO)
        notes = release.extract_notes(out, "9.5.0")
        assert "### Removed" not in notes
        assert "### Added" in notes
        assert "A nasty bug (#124)" in notes


class TestExtractNotes:
    def test_extracts_requested_section_only(self):
        notes = release.extract_notes(SAMPLE, "9.4.0")
        assert "Older thing" in notes
        assert "A shiny new feature" not in notes

    def test_missing_version_raises(self):
        with pytest.raises(release.ReleaseError):
            release.extract_notes(SAMPLE, "1.2.3")



"""Define tests for weaather module."""

import pyatmo
from pyatmo import DeviceType
from pyatmo.modules.base_class import Location, Place


async def test_async_weather_NAMain(async_home):
    """Test Netatmo weather station main module."""
    module_id = "12:34:56:80:bb:26"
    assert module_id in async_home.modules
    module = async_home.modules[module_id]
    assert module.device_type == DeviceType.NAMain


async def test_async_weather_update_main(async_account):
    """Test main weather station module."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_weather_stations()
    home = async_account.homes[home_id]

    module_id = "12:34:56:80:bb:26"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.device_type == DeviceType.NAMain
    assert module.name == "Villa"
    assert module.modules == [
        "12:34:56:80:44:92",
        "12:34:56:80:7e:18",
        "12:34:56:80:1c:42",
        "12:34:56:80:c1:ea",
    ]
    assert module.features == {
        "temperature",
        "humidity",
        "co2",
        "noise",
        "pressure",
        "absolute_pressure",
        "temp_trend",
        "pressure_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "reachable",
        "wifi_strength",
        "place",
    }
    assert module.firmware_revision == 181
    assert module.wifi_strength == 57
    assert module.temperature == 21.1
    assert module.humidity == 45
    assert module.co2 == 1339
    assert module.pressure == 1026.8
    assert module.noise == 35
    assert module.absolute_pressure == 974.5
    assert module.place == Place(
        {
            "altitude": 329,
            "city": "Someplace",
            "country": "FR",
            "location": Location(longitude=6.1234567, latitude=46.123456),
            "timezone": "Europe/Paris",
        },
    )


async def test_async_weather_update_indoor(async_account):
    """Test indoor weather module."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_weather_stations()
    home = async_account.homes[home_id]

    module_id = "12:34:56:80:44:92"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.name == "Villa Bedroom"
    assert module.features == {
        "temperature",
        "temp_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "reachable",
        "rf_strength",
        "co2",
        "humidity",
        "battery",
        "place",
    }
    assert module.device_type == DeviceType.NAModule4
    assert module.modules is None
    assert module.firmware_revision == 51
    assert module.rf_strength == 67
    assert module.temperature == 19.3
    assert module.humidity == 53
    assert module.battery == 28


async def test_async_weather_update_rain(async_account):
    """Test rain module."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_weather_stations()
    home = async_account.homes[home_id]

    module_id = "12:34:56:80:c1:ea"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.name == "Villa Rain"
    assert module.features == {
        "sum_rain_1",
        "sum_rain_24",
        "rain",
        "reachable",
        "rf_strength",
        "battery",
        "place",
    }
    assert module.device_type == DeviceType.NAModule3
    assert module.modules is None
    assert module.firmware_revision == 12
    assert module.rf_strength == 79
    assert module.rain == 3.7


async def test_async_weather_update_outdoor(async_account):
    """Test outdoor module."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_weather_stations()
    home = async_account.homes[home_id]

    module_id = "12:34:56:80:1c:42"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.name == "Villa Outdoor"
    assert module.features == {
        "temperature",
        "humidity",
        "temp_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "reachable",
        "rf_strength",
        "battery",
        "place",
    }
    assert module.device_type == DeviceType.NAModule1
    assert module.modules is None
    assert module.firmware_revision == 50
    assert module.rf_strength == 68
    assert module.reachable is False


async def test_async_weather_update_wind(async_account):
    """Test wind module."""
    home_id = "91763b24c43d3e344f424e8b"
    await async_account.async_update_weather_stations()
    home = async_account.homes[home_id]

    module_id = "12:34:56:03:1b:e4"
    assert module_id in home.modules
    module = home.modules[module_id]
    assert module.name == "Villa Garden"
    assert module.features == {
        "wind_strength",
        "gust_strength",
        "gust_angle",
        "gust_direction",
        "wind_angle",
        "wind_direction",
        "reachable",
        "rf_strength",
        "battery",
        "place",
    }
    assert module.device_type == DeviceType.NAModule2
    assert module.modules is None
    assert module.firmware_revision == 19
    assert module.rf_strength == 59
    assert module.wind_strength == 4
    assert module.wind_angle == 217
    assert module.gust_strength == 9
    assert module.gust_angle == 206


async def test_async_weather_favorite(async_account):
    """Test favorite weather station."""
    await async_account.async_update_weather_stations()

    module_id = "00:11:22:2c:be:c8"
    assert module_id in async_account.modules
    module = async_account.modules[module_id]
    assert module.device_type == DeviceType.NAMain
    assert module.name == "Zuhause (Kinderzimmer)"
    assert module.modules == ["00:11:22:2c:ce:b6"]
    assert module.features == {
        "temperature",
        "humidity",
        "co2",
        "noise",
        "pressure",
        "absolute_pressure",
        "temp_trend",
        "pressure_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "reachable",
        "wifi_strength",
        "place",
    }
    assert module.pressure == 1015.6
    assert module.absolute_pressure == 1000.4
    assert module.place == Place(
        {
            "altitude": 127,
            "city": "Wiesbaden",
            "country": "DE",
            "location": Location(
                longitude=8.238054275512695,
                latitude=50.07585525512695,
            ),
            "timezone": "Europe/Berlin",
        },
    )

    module_id = "00:11:22:2c:ce:b6"
    assert module_id in async_account.modules
    module = async_account.modules[module_id]
    assert module.device_type == DeviceType.NAModule1
    assert module.name == "Unknown"
    assert module.modules is None
    assert module.features == {
        "temperature",
        "humidity",
        "temp_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "reachable",
        "rf_strength",
        "battery",
        "place",
    }
    assert module.temperature == 7.8
    assert module.humidity == 87


async def test_async_air_care_update(async_account):
    """Test basic air care update."""
    await async_account.async_update_air_care()

    module_id = "12:34:56:26:68:92"
    assert module_id in async_account.modules
    module = async_account.modules[module_id]

    assert module.device_type == DeviceType.NHC
    assert module.name == "Baby Bedroom"
    assert module.features == {
        "temperature",
        "humidity",
        "co2",
        "noise",
        "pressure",
        "absolute_pressure",
        "temp_trend",
        "pressure_trend",
        "min_temp",
        "max_temp",
        "temp_max",
        "temp_min",
        "health_idx",
        "reachable",
        "wifi_strength",
        "place",
    }

    assert module.modules is None
    assert module.firmware_revision == 45
    assert module.wifi_strength == 68
    assert module.temperature == 21.6
    assert module.humidity == 66
    assert module.co2 == 1053
    assert module.pressure == 1021.4
    assert module.noise == 45
    assert module.absolute_pressure == 1011
    assert module.health_idx == 1


async def test_register_public_weather_area_generates_unique_ids(async_account):
    """Test that registering areas without area_id yields unique ids."""
    lon_ne = "6.221652"
    lat_ne = "46.610870"
    lon_sw = "6.217828"
    lat_sw = "46.596485"

    area_id_1 = async_account.register_public_weather_area(
        lat_ne,
        lon_ne,
        lat_sw,
        lon_sw,
    )
    area_id_2 = async_account.register_public_weather_area(
        lat_ne,
        lon_ne,
        lat_sw,
        lon_sw,
    )

    assert area_id_1 != area_id_2
    assert len(async_account.public_weather_areas) == 2


async def test_async_public_weather_update(async_account):
    """Test basic public weather update."""
    lon_ne = "6.221652"
    lat_ne = "46.610870"
    lon_sw = "6.217828"
    lat_sw = "46.596485"

    area_id = async_account.register_public_weather_area(lat_ne, lon_ne, lat_sw, lon_sw)
    await async_account.async_update_public_weather(area_id)

    area = async_account.public_weather_areas[area_id]
    assert area.location == pyatmo.modules.Location(
        lat_ne,
        lon_ne,
        lat_sw,
        lon_sw,
    )
    assert area.stations_in_area() == 8

    assert area.get_latest_rain() == {
        "70:ee:50:1f:68:9e": 0,
        "70:ee:50:27:25:b0": 0,
        "70:ee:50:36:94:7c": 0.5,
        "70:ee:50:36:a9:fc": 0,
    }

    assert area.get_60_min_rain() == {
        "70:ee:50:1f:68:9e": 0,
        "70:ee:50:27:25:b0": 0,
        "70:ee:50:36:94:7c": 0.2,
        "70:ee:50:36:a9:fc": 0,
    }

    assert area.get_24_h_rain() == {
        "70:ee:50:1f:68:9e": 9.999,
        "70:ee:50:27:25:b0": 11.716000000000001,
        "70:ee:50:36:94:7c": 12.322000000000001,
        "70:ee:50:36:a9:fc": 11.009,
    }

    assert area.get_latest_pressures() == {
        "70:ee:50:1f:68:9e": 1007.3,
        "70:ee:50:27:25:b0": 1012.8,
        "70:ee:50:36:94:7c": 1010.6,
        "70:ee:50:36:a9:fc": 1010,
        "70:ee:50:01:20:fa": 1014.4,
        "70:ee:50:04:ed:7a": 1005.4,
        "70:ee:50:27:9f:2c": 1010.6,
        "70:ee:50:3c:02:78": 1011.7,
    }

    assert area.get_latest_temperatures() == {
        "70:ee:50:1f:68:9e": 21.1,
        "70:ee:50:27:25:b0": 23.2,
        "70:ee:50:36:94:7c": 21.4,
        "70:ee:50:36:a9:fc": 20.1,
        "70:ee:50:01:20:fa": 27.4,
        "70:ee:50:04:ed:7a": 19.8,
        "70:ee:50:27:9f:2c": 25.5,
        "70:ee:50:3c:02:78": 23.3,
    }

    assert area.get_latest_humidities() == {
        "70:ee:50:1f:68:9e": 69,
        "70:ee:50:27:25:b0": 60,
        "70:ee:50:36:94:7c": 62,
        "70:ee:50:36:a9:fc": 67,
        "70:ee:50:01:20:fa": 58,
        "70:ee:50:04:ed:7a": 76,
        "70:ee:50:27:9f:2c": 56,
        "70:ee:50:3c:02:78": 58,
    }

    assert area.get_latest_wind_strengths() == {"70:ee:50:36:a9:fc": 15}

    assert area.get_latest_wind_angles() == {"70:ee:50:36:a9:fc": 17}

    assert area.get_latest_gust_strengths() == {"70:ee:50:36:a9:fc": 31}

    assert area.get_latest_gust_angles() == {"70:ee:50:36:a9:fc": 217}



"""Module to represent a Netatmo person."""

from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING

from pyatmo.modules.base_class import NetatmoBase

if TYPE_CHECKING:
    from pyatmo.const import RawData

    from .home import Home

LOG: logging.Logger = logging.getLogger(__name__)


@dataclass
class Person(NetatmoBase):
    """Class to represent a Netatmo person."""

    pseudo: str | None
    url: str | None
    out_of_sight: bool | None = None
    last_seen: int | None = None

    def __init__(self, home: Home, raw_data: RawData) -> None:
        """Initialize a Netatmo person instance."""

        super().__init__(raw_data)
        self.home = home
        self.pseudo = raw_data.get("pseudo")
        self.url = raw_data.get("url")

    def update(self, raw_data: RawData) -> None:
        """Update person data."""
        self.out_of_sight = raw_data.get("out_of_sight")
        self.last_seen = raw_data.get("last_seen")
